"""
Adapted from: https://github.com/thu-ml/tianshou/blob/master/tianshou/utils/net/common.py
"""

from typing import Any, Dict, List, Optional, Tuple, Type, Union

import torch
from torch import Tensor, nn

ModuleType = Optional[Type[nn.Module]]
ArgType = Union[Tuple[Any, ...], Dict[Any, Any], None]
ArgsType = Union[ArgType, List[ArgType]]


def create_layer_with_args(layer_type: ModuleType, layer_args: Optional[ArgType]) -> nn.Module:
    """Create a single layer with given layer type and arguments.

    Args:
        layer_type (ModuleType): the type of the layer to be created.
        layer_args (ArgType, optional): the arguments to be passed to the layer.
    """
    if layer_type is None:
        raise ValueError("`layer_type` must be not None")
    if isinstance(layer_args, tuple):
        return layer_type(*layer_args)
    elif isinstance(layer_args, dict):
        return layer_type(**layer_args)
    elif layer_args is None:
        return layer_type()
    else:
        raise ValueError(f"`layer_args` must be None, tuple or dict, got {type(layer_args)}")


def miniblock(
    input_size: int,
    output_size: int,
    layer_type: Type[nn.Module] = nn.Linear,
    layer_args: ArgType = None,
    dropout_layer: ModuleType = None,
    dropout_args: ArgType = None,
    norm_layer: ModuleType = None,
    norm_args: ArgType = None,
    activation: ModuleType = None,
    act_args: ArgType = None,
) -> List[nn.Module]:
    """Construct a miniblock with given input/output-size, dropout layer, norm layer and activation function.

    Based on Tianshou's miniblock function
    (https://github.com/thu-ml/tianshou/blob/master/tianshou/utils/net/common.py).

    Args:
        input_size (int): the input size of the miniblock (in_features for Linear and in_channels for Conv2d).
        output_size (int): the output size of the miniblock.
        layer_type (Type[nn.Linear], optional): the type of the layer to be created. Defaults to nn.Linear.
        layer_args (ArgType, optional): the arguments to be passed to the layer.
            Defaults to None.
        dropout_layer (ModuleType, optional): the type of the dropout layer to be created. Defaults to None.
        dropout_args (ArgType, optional): the arguments to be passed to the dropout
            layer. Defaults to None.
        norm_layer (ModuleType, optional): the type of the norm layer to be created. Defaults to None.
        norm_args (ArgType, optional): the arguments to be passed to the norm layer.
            Defaults to None.
        activation (ModuleType, optional): the type of the activation function to be created.
            Defaults to None.
        act_args (Tuple[Any, ...] | Dict[Any, Any] | None, optional): the arguments to be passed to the activation
            function. Defaults to None.

    Returns:
        List[nn.Module]: the miniblock as a list of layers.
    """
    if layer_args is None:
        layers: List[nn.Module] = [layer_type(input_size, output_size)]
    elif isinstance(layer_args, tuple):
        layers = [layer_type(input_size, output_size, *layer_args)]
    elif isinstance(layer_args, dict):
        layers = [layer_type(input_size, output_size, **layer_args)]
    else:
        raise ValueError(f"layer_args must be None, tuple or dict, got {type(layer_args)}")

    if dropout_layer is not None:
        layers += [create_layer_with_args(dropout_layer, dropout_args)]

    if norm_layer is not None:
        layers += [create_layer_with_args(norm_layer, norm_args)]

    if activation is not None:
        layers += [create_layer_with_args(activation, act_args)]
    return layers


def create_layers(
    layer_type: Union[ModuleType, List[ModuleType]], layer_args: Optional[ArgsType], num_layers: int
) -> Tuple[List[ModuleType], ArgsType]:
    """Create a list of layers with given layer type and arguments.

    If a layer_type is not specified, then the lists will be filled with None. If the layer type or the layer arguments
    are specified only once, they will be cast to a sequence of length num_layers.

    Args:
        layer_type (Union[ModuleType, Sequence[ModuleType]]): the type of the layer to be created.
        layer_args (ArgsType, optional): the arguments to be passed to the layer.
        num_layers (int): the number of layers to be created.

    Returns:
        Tuple[Sequence[ModuleType], ArgsType]: a list of layers and a list of args.

    Examples:
        >>> create_layers(nn.Linear, None, 3)
        ([nn.Linear, nn.Linear, nn.Linear], [None, None, None])

        >>> create_layers(nn.Linear, {"arg1":3, "arg2": "foo"}, 3)
        (
            [nn.Linear, nn.Linear, nn.Linear],
            [{'arg1': 3, 'arg2': 'foo'}, {'arg1': 3, 'arg2': 'foo'}, {'arg1': 3, 'arg2': 'foo'}]
        )

        >>> create_layers([nn.Linear, nn.Conv2d], [{"bias":False}, {"kernel_size": 5, "bias": True}], 2)
        ([nn.Linear, nn.Conv2d], [{'bias': False}, {'kernel_size':5, 'bias': True}])

        >>> create_layers([nn.Linear, nn.Linear], (64, 10), 2)
        ([nn.Linear, nn.Linear], [(64, 10), (64, 10)])
    """
    if layer_type is None:
        layers_list = [None] * num_layers
        args_list = [None] * num_layers
        return layers_list, args_list

    if isinstance(layer_type, list):
        assert len(layer_type) == num_layers
        layers_list = layer_type
        if isinstance(layer_args, list):
            assert len(layer_args) == num_layers
            args_list = layer_args
        else:
            args_list = [layer_args for _ in range(num_layers)]
    else:
        layers_list = [layer_type for _ in range(num_layers)]
        args_list = [layer_args for _ in range(num_layers)]
    return layers_list, args_list


def per_layer_ortho_init_weights(module: nn.Module, gain: float = 1.0, bias: float = 0.0):
    """Initialize the weights of a module with orthogonal weights.

    Args:
        module (nn.Module): module to initialize
        gain (float, optional): gain of the orthogonal initialization. Defaults to 1.0.
        bias (float, optional): bias of the orthogonal initialization. Defaults to 0.0.
    """
    if isinstance(module, nn.Linear):
        nn.init.orthogonal_(module.weight, gain=gain)
        if module.bias is not None:
            module.bias.data.fill_(bias)
    elif isinstance(module, nn.LSTM):
        for name, param in module.named_parameters():
            if "bias" in name:
                nn.init.constant_(param, val=bias)
            elif "weight" in name:
                nn.init.orthogonal_(param, gain=gain)
    elif isinstance(module, (nn.Sequential, nn.ModuleList)):
        for i in range(len(module)):
            per_layer_ortho_init_weights(module[i], gain=gain, bias=bias)


def cnn_forward(
    model: nn.Module,
    input: Tensor,
    input_dim: Union[torch.Size, Tuple[int, ...]],
    output_dim: Union[torch.Size, Tuple[int, ...]],
) -> Tensor:
    """
    Compute the forward of a Convolutional neural network.
    It flattens all the dimensions before the model input_size, i.e.,
    the dimensions before the (C_in, H, W) dimensions for the encoder
    and the dimensions before the (feature_size,) dimension for the decoder.

    Args:
        model (nn.Module): the model.
        input (Tensor): the input tensor of dimension (*, C_in, H, W) or (*, feature_size),
            where * means any number of dimensions including None.
        input_dim (Union[torch.Size, Tuple[int, ...]]): the input dimensions,
            i.e., either (C_in, H, W) or (feature_size,).
        output_dim (Union[torch.Size, Tuple[int, ...]]): the desired dimensions in output.

    Returns:
        The output of dimensions (*, *output_dim).

    Examples:
        >>> encoder
        CNN(
            (network): Sequential(
                (0): Conv2d(3, 4, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
                (1): ReLU()
                (2): Conv2d(4, 8, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
                (3): ReLU()
                (4): Flatten(start_dim=1, end_dim=-1)
                (5): Linear(in_features=128, out_features=25, bias=True)
            )
        )
        >>> input = torch.rand(10, 20, 3, 4, 4)
        >>> cnn_forward(encoder, input, (3, 4, 4), -1).shape
        torch.Size([10, 20, 25])

        >>> decoder
        Sequential(
            (0): Linear(in_features=230, out_features=1024, bias=True)
            (1): Unflatten(dim=-1, unflattened_size=(1024, 1, 1))
            (2): ConvTranspose2d(1024, 128, kernel_size=(5, 5), stride=(2, 2))
            (3): ReLU()
            (4): ConvTranspose2d(128, 64, kernel_size=(5, 5), stride=(2, 2))
            (5): ReLU()
            (6): ConvTranspose2d(64, 32, kernel_size=(6, 6), stride=(2, 2))
            (7): ReLU()
            (8): ConvTranspose2d(32, 3, kernel_size=(6, 6), stride=(2, 2))
        )
        >>> input = torch.rand(10, 20, 230)
        >>> cnn_forward(decoder, input, (230,), (3, 64, 64)).shape
        torch.Size([10, 20, 3, 64, 64])
    """
    batch_shapes = input.shape[: -len(input_dim)]
    flatten_input = input.reshape(-1, *input_dim)
    model_out = model(flatten_input)
    return model_out.reshape(*batch_shapes, *output_dim)


"""
Adapted from: https://github.com/thu-ml/tianshou/blob/master/tianshou/utils/net/common.py
"""

import warnings
from math import prod
from typing import Any, Callable, Dict, Optional, Sequence, Union, no_type_check

import torch
import torch.nn.functional as F
from torch import Tensor, nn


class MLP(nn.Module):
    """Simple MLP backbone.

    Args:
        input_dims (Union[int, Sequence[int]]): dimensions of the input vector.
        output_dim (int, optional): dimension of the output vector. If set to None, there
            is no final linear layer. Else, a final linear layer is added.
            Defaults to None.
        hidden_sizes (Sequence[int], optional): shape of MLP passed in as a list, not including
            input_dims and output_dim.
        dropout_layer (Union[ModuleType, Sequence[ModuleType]], optional): which dropout layer to be used
            before activation (possibly before the normalization layer), e.g., ``nn.Dropout``.
            You can also pass a list of dropout modules with the same length
            of hidden_sizes to use different dropout modules in different layers.
            If None, then no dropout layer is used.
            Defaults to None.
        norm_layer (Union[ModuleType, Sequence[ModuleType]], optional): which normalization layer to be used
            before activation, e.g., ``nn.LayerNorm`` and ``nn.BatchNorm1d``.
            You can also pass a list of normalization modules with the same length
            of hidden_sizes to use different normalization modules in different layers.
            If None, then no normalization layer is used.
            Defaults to None.
        activation (Union[ModuleType, Sequence[ModuleType]], optional): which activation to use after each layer,
            can be both the same activation for all layers if a single ``nn.Module`` is passed, or different
            activations for different layers if a list is passed.
            Defaults to ``nn.ReLU``.
        flatten_dim (int, optional): whether to flatten input data. The flatten dimension starts from 1.
            Defaults to True.
    """

    def __init__(
        self,
        input_dims: Union[int, Sequence[int]],
        output_dim: Optional[int] = None,
        hidden_sizes: Sequence[int] = (),
        layer_args: Optional[ArgsType] = None,
        dropout_layer: Optional[Union[ModuleType, Sequence[ModuleType]]] = None,
        dropout_args: Optional[ArgsType] = None,
        norm_layer: Optional[Union[ModuleType, Sequence[ModuleType]]] = None,
        norm_args: Optional[ArgsType] = None,
        activation: Optional[Union[ModuleType, Sequence[ModuleType]]] = nn.ReLU,
        act_args: Optional[ArgsType] = None,
        flatten_dim: Optional[int] = None,
    ) -> None:
        super().__init__()
        num_layers = len(hidden_sizes)
        if num_layers < 1 and output_dim is None:
            raise ValueError("The number of layers should be at least 1.")

        if isinstance(input_dims, Sequence) and flatten_dim is None:
            warnings.warn(
                "input_dims is a sequence, but flatten_dim is not specified. "
                "Be careful to flatten the input data correctly before the forward."
            )

        dropout_layer_list, dropout_args_list = create_layers(dropout_layer, dropout_args, num_layers)
        norm_layer_list, norm_args_list = create_layers(norm_layer, norm_args, num_layers)
        activation_list, act_args_list = create_layers(activation, act_args, num_layers)

        if isinstance(layer_args, list):
            layer_args_list = layer_args
        else:
            layer_args_list = [layer_args] * num_layers

        if isinstance(input_dims, int):
            input_dims = [input_dims]
        hidden_sizes = [prod(input_dims)] + list(hidden_sizes)
        model = []
        for in_dim, out_dim, l_args, drop, drop_args, norm, norm_args, activ, act_args in zip(
            hidden_sizes[:-1],
            hidden_sizes[1:],
            layer_args_list,
            dropout_layer_list,
            dropout_args_list,
            norm_layer_list,
            norm_args_list,
            activation_list,
            act_args_list,
        ):
            model += miniblock(in_dim, out_dim, nn.Linear, l_args, drop, drop_args, norm, norm_args, activ, act_args)
        if output_dim is not None:
            model += [nn.Linear(hidden_sizes[-1], output_dim)]

        self._output_dim = output_dim or hidden_sizes[-1]
        self._model = nn.Sequential(*model)
        self._flatten_dim = flatten_dim

    @property
    def model(self) -> nn.Module:
        return self._model

    @property
    def output_dim(self) -> int:
        return self._output_dim

    @property
    def flatten_dim(self) -> Optional[int]:
        return self._flatten_dim

    @no_type_check
    def forward(self, obs: Tensor) -> Tensor:
        if self.flatten_dim is not None:
            obs = obs.flatten(self.flatten_dim)
        return self.model(obs)


class CNN(nn.Module):
    """Simple CNN backbone.

    Args:
        input_channels (int): dimensions of the input channels.
        hidden_channels (Sequence[int], optional): intermediate number of channels for the CNN,
            including the output channels.
        dropout_layer (Union[ModuleType, Sequence[ModuleType]], optional): which dropout layer to be used
            before activation (possibly before the normalization layer), e.g., ``nn.Dropout``.
            You can also pass a list of dropout modules with the same length
            of hidden_sizes to use different dropout modules in different layers.
            If None, then no dropout layer is used.
            Defaults to None.
        norm_layer (Union[ModuleType, Sequence[ModuleType]], optional): which normalization layer to be used
            before activation, e.g., ``nn.LayerNorm`` and ``nn.BatchNorm1d``.
            You can also pass a list of normalization modules with the same length
            of hidden_sizes to use different normalization modules in different layers.
            If None, then no normalization layer is used.
            Defaults to None.
        activation (Union[ModuleType, Sequence[ModuleType]], optional): which activation to use after each layer,
            can be both the same activation for all layers if a single ``nn.Module`` is passed, or different
            activations for different layers if a list is passed.
            Defaults to ``nn.ReLU``.
    """

    def __init__(
        self,
        input_channels: int,
        hidden_channels: Sequence[int],
        cnn_layer: ModuleType = nn.Conv2d,
        layer_args: ArgsType = None,
        dropout_layer: Optional[Union[ModuleType, Sequence[ModuleType]]] = None,
        dropout_args: Optional[ArgsType] = None,
        norm_layer: Optional[Union[ModuleType, Sequence[ModuleType]]] = None,
        norm_args: Optional[ArgsType] = None,
        activation: Optional[Union[ModuleType, Sequence[ModuleType]]] = nn.ReLU,
        act_args: Optional[ArgsType] = None,
    ) -> None:
        super().__init__()
        num_layers = len(hidden_channels)
        if num_layers < 1:
            raise ValueError("The number of layers should be at least 1.")

        dropout_layer_list, dropout_args_list = create_layers(dropout_layer, dropout_args, num_layers)
        norm_layer_list, norm_args_list = create_layers(norm_layer, norm_args, num_layers)
        activation_list, act_args_list = create_layers(activation, act_args, num_layers)

        if isinstance(layer_args, list):
            layer_args_list = layer_args
        else:
            layer_args_list = [layer_args] * num_layers

        hidden_sizes = [input_channels] + list(hidden_channels)
        model = []
        for in_dim, out_dim, l_args, drop, drop_args, norm, norm_args, activ, act_args in zip(
            hidden_sizes[:-1],
            hidden_sizes[1:],
            layer_args_list,
            dropout_layer_list,
            dropout_args_list,
            norm_layer_list,
            norm_args_list,
            activation_list,
            act_args_list,
        ):
            model += miniblock(in_dim, out_dim, cnn_layer, l_args, drop, drop_args, norm, norm_args, activ, act_args)

        self._output_dim = hidden_sizes[-1]
        self._model = nn.Sequential(*model)

    @property
    def model(self) -> nn.Module:
        return self._model

    @property
    def output_dim(self) -> int:
        return self._output_dim

    @no_type_check
    def forward(self, obs: Tensor) -> Tensor:
        return self.model(obs)


class DeCNN(nn.Module):
    """Simple DeCNN backbone.

    Args:
        input_channels (int): dimensions of the input channels.
        hidden_channels (Sequence[int], optional): intermediate number of channels for the CNN,
            including the output channels.
        dropout_layer (Union[ModuleType, Sequence[ModuleType]], optional): which dropout layer to be used
            before activation (possibly before the normalization layer), e.g., ``nn.Dropout``.
            You can also pass a list of dropout modules with the same length
            of hidden_sizes to use different dropout modules in different layers.
            If None, then no dropout layer is used.
            Defaults to None.
        norm_layer (Union[ModuleType, Sequence[ModuleType]], optional): which normalization layer to be used
            before activation, e.g., ``nn.LayerNorm`` and ``nn.BatchNorm1d``.
            You can also pass a list of normalization modules with the same length
            of hidden_sizes to use different normalization modules in different layers.
            If None, then no normalization layer is used.
            Defaults to None.
        activation (Union[ModuleType, Sequence[ModuleType]], optional): which activation to use after each layer,
            can be both the same activation for all layers if a single ``nn.Module`` is passed, or different
            activations for different layers if a list is passed.
            Defaults to ``nn.ReLU``.
    """

    def __init__(
        self,
        input_channels: int,
        hidden_channels: Sequence[int] = (),
        cnn_layer: ModuleType = nn.ConvTranspose2d,
        layer_args: ArgsType = None,
        dropout_layer: Optional[Union[ModuleType, Sequence[ModuleType]]] = None,
        dropout_args: Optional[ArgsType] = None,
        norm_layer: Optional[Union[ModuleType, Sequence[ModuleType]]] = None,
        norm_args: Optional[ArgsType] = None,
        activation: Optional[Union[ModuleType, Sequence[ModuleType]]] = nn.ReLU,
        act_args: Optional[ArgsType] = None,
    ) -> None:
        super().__init__()
        num_layers = len(hidden_channels)
        if num_layers < 1:
            raise ValueError("The number of layers should be at least 1.")

        dropout_layer_list, dropout_args_list = create_layers(dropout_layer, dropout_args, num_layers)
        norm_layer_list, norm_args_list = create_layers(norm_layer, norm_args, num_layers)
        activation_list, act_args_list = create_layers(activation, act_args, num_layers)

        if isinstance(layer_args, list):
            layer_args_list = layer_args
        else:
            layer_args_list = [layer_args] * num_layers

        hidden_sizes = [input_channels] + list(hidden_channels)
        model = []
        for in_dim, out_dim, l_args, drop, drop_args, norm, norm_args, activ, act_args in zip(
            hidden_sizes[:-1],
            hidden_sizes[1:],
            layer_args_list,
            dropout_layer_list,
            dropout_args_list,
            norm_layer_list,
            norm_args_list,
            activation_list,
            act_args_list,
        ):
            model += miniblock(in_dim, out_dim, cnn_layer, l_args, drop, drop_args, norm, norm_args, activ, act_args)

        self._output_dim = hidden_sizes[-1]
        self._model = nn.Sequential(*model)

    @property
    def model(self) -> nn.Module:
        return self._model

    @property
    def output_dim(self) -> int:
        return self._output_dim

    @no_type_check
    def forward(self, obs: Tensor) -> Tensor:
        return self.model(obs)


class NatureCNN(CNN):
    """CNN from DQN Nature paper: Mnih, Volodymyr, et al. "Human-level control through deep reinforcement learning."
    Nature 518.7540 (2015): 529-533.

    Args:
        in_channels (int): the input channels to the first convolutional layer
        features_dim (int): the features dimension in output from the last convolutional layer
        screen_size (int, optional): the dimension of the input image as a single integer.
            Needed to extract the features and compute the output dimension after all the
            convolutional layers.
            Defaults to 64.
    """

    def __init__(self, in_channels: int, features_dim: int, screen_size: int = 64):
        super().__init__(
            in_channels,
            [32, 64, 64],
            layer_args=[
                {"kernel_size": 8, "stride": 4},
                {"kernel_size": 4, "stride": 2},
                {"kernel_size": 3, "stride": 1},
            ],
        )

        with torch.no_grad():
            x = self.model(torch.rand(1, in_channels, screen_size, screen_size, device=self.model[0].weight.device))
            out_dim = x.flatten(1).shape[1]
        self._output_dim = out_dim
        self.fc = None
        if features_dim is not None:
            self._output_dim = features_dim
            self.fc = nn.Linear(out_dim, features_dim)

    @property
    def output_dim(self) -> int:
        return self._output_dim

    def forward(self, x: Tensor) -> Tensor:
        x = cnn_forward(self.model, x, input_dim=x.shape[-3:], output_dim=(-1,))
        x = F.relu(self.fc(x))
        return x


class LayerNormGRUCell(nn.Module):
    """A GRU cell with a LayerNorm, taken
    from https://github.com/danijar/dreamerv2/blob/main/dreamerv2/common/nets.py#L317.

    This particular GRU cell accepts 3-D inputs, with a sequence of length 1, and applies
    a LayerNorm after the projection of the inputs.

    Args:
        input_size (int): the input size.
        hidden_size (int): the hidden state size
        bias (bool, optional): whether to apply a bias to the input projection.
            Defaults to True.
        batch_first (bool, optional): whether the first dimension represent the batch dimension or not.
            Defaults to False.
        layer_norm_cls (Callable[..., nn.Module]): the layer norm to apply after the input projection.
            Defaults to nn.Identiy.
        layer_norm_kw (Dict[str, Any]): the kwargs of the layer norm.
            Default to {}.
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        bias: bool = True,
        batch_first: bool = False,
        layer_norm_cls: Callable[..., nn.Module] = nn.Identity,
        layer_norm_kw: Dict[str, Any] = {},
    ) -> None:
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.bias = bias
        self.batch_first = batch_first
        self.linear = nn.Linear(input_size + hidden_size, 3 * hidden_size, bias=self.bias)
        # Avoid multiple values for the `normalized_shape` argument
        layer_norm_kw.pop("normalized_shape", None)
        self.layer_norm = layer_norm_cls(3 * hidden_size, **layer_norm_kw)

    def forward(self, input: Tensor, hx: Optional[Tensor] = None) -> Tensor:
        is_3d = input.dim() == 3
        if is_3d:
            if input.shape[int(self.batch_first)] == 1:
                input = input.squeeze(int(self.batch_first))
            else:
                raise AssertionError(
                    "LayerNormGRUCell: Expected input to be 3-D with sequence length equal to 1 but received "
                    f"a sequence of length {input.shape[int(self.batch_first)]}"
                )
        if hx.dim() == 3:
            hx = hx.squeeze(0)
        assert input.dim() in (
            1,
            2,
        ), f"LayerNormGRUCell: Expected input to be 1-D or 2-D but received {input.dim()}-D tensor"

        is_batched = input.dim() == 2
        if not is_batched:
            input = input.unsqueeze(0)

        if hx is None:
            hx = torch.zeros(input.size(0), self.hidden_size, dtype=input.dtype, device=input.device)
        else:
            hx = hx.unsqueeze(0) if not is_batched else hx

        input = torch.cat((hx, input), -1)
        x = self.linear(input)
        x = self.layer_norm(x)
        reset, cand, update = torch.chunk(x, 3, -1)
        reset = torch.sigmoid(reset)
        cand = torch.tanh(reset * cand)
        update = torch.sigmoid(update - 1)
        hx = update * cand + (1 - update) * hx

        if not is_batched:
            hx = hx.squeeze(0)
        elif is_3d:
            hx = hx.unsqueeze(0)

        return hx


class MultiEncoder(nn.Module):
    def __init__(
        self,
        cnn_encoder: ModuleType,
        mlp_encoder: ModuleType,
    ) -> None:
        super().__init__()
        if cnn_encoder is None and mlp_encoder is None:
            raise ValueError("There must be at least one encoder, both cnn and mlp encoders are None")
        self.has_cnn_encoder = False
        self.has_mlp_encoder = False
        if cnn_encoder is not None:
            if getattr(cnn_encoder, "input_dim", None) is None:
                raise AttributeError(
                    "`cnn_encoder` must contain the `input_dim` attribute representing "
                    "the dimension of the input tensor"
                )
            if getattr(cnn_encoder, "output_dim", None) is None:
                raise AttributeError(
                    "`cnn_encoder` must contain the `output_dim` attribute representing "
                    "the dimension of the output tensor"
                )
            self.has_cnn_encoder = True
        if mlp_encoder is not None:
            if getattr(mlp_encoder, "input_dim", None) is None:
                raise AttributeError(
                    "`mlp_encoder` must contain the `input_dim` attribute representing "
                    "the dimension of the input tensor"
                )
            if getattr(mlp_encoder, "output_dim", None) is None:
                raise AttributeError(
                    "`mlp_encoder` must contain the `output_dim` attribute representing "
                    "the dimension of the output tensor"
                )
            self.has_mlp_encoder = True
        self.has_both_encoders = self.has_cnn_encoder and self.has_mlp_encoder
        self.cnn_encoder = cnn_encoder
        self.mlp_encoder = mlp_encoder
        self.cnn_input_dim = self.cnn_encoder.input_dim if self.cnn_encoder is not None else None
        self.mlp_input_dim = self.mlp_encoder.input_dim if self.mlp_encoder is not None else None
        self.cnn_output_dim = self.cnn_encoder.output_dim if self.cnn_encoder is not None else 0
        self.mlp_output_dim = self.mlp_encoder.output_dim if self.mlp_encoder is not None else 0
        self.output_dim = self.cnn_output_dim + self.mlp_output_dim

    @property
    def cnn_keys(self) -> Sequence[str]:
        return self.cnn_encoder.keys if self.cnn_encoder is not None else []

    @property
    def mlp_keys(self) -> Sequence[str]:
        return self.mlp_encoder.keys if self.mlp_encoder is not None else []

    def forward(self, obs: Dict[str, Tensor], *args, **kwargs) -> Tensor:
        if self.has_cnn_encoder:
            cnn_out = self.cnn_encoder(obs, *args, **kwargs)
        if self.has_mlp_encoder:
            mlp_out = self.mlp_encoder(obs, *args, **kwargs)
        if self.has_both_encoders:
            return torch.cat((cnn_out, mlp_out), -1)
        elif self.has_cnn_encoder:
            return cnn_out
        else:
            return mlp_out


class MultiDecoder(nn.Module):
    def __init__(
        self,
        cnn_decoder: ModuleType,
        mlp_decoder: ModuleType,
    ) -> None:
        super().__init__()
        if cnn_decoder is None and mlp_decoder is None:
            raise ValueError("There must be an decoder, both cnn and mlp decoders are None")
        self.cnn_decoder = cnn_decoder
        self.mlp_decoder = mlp_decoder

    @property
    def cnn_keys(self) -> Sequence[str]:
        return self.cnn_decoder.keys if self.cnn_decoder is not None else []

    @property
    def mlp_keys(self) -> Sequence[str]:
        return self.mlp_decoder.keys if self.mlp_decoder is not None else []

    # adapt to xuance: the return Dict[str, Tensor] modified to -> Tensor
    def forward(self, x: Tensor) -> Tensor:
        reconstructed_obs = None
        if self.cnn_decoder is not None:
            reconstructed_obs = self.cnn_decoder(x)[0]
        if self.mlp_decoder is not None:
            reconstructed_obs = self.mlp_decoder(x)
        return reconstructed_obs


class LayerNormChannelLast(nn.LayerNorm):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def forward(self, x: Tensor) -> Tensor:
        if x.dim() != 4:
            raise ValueError(f"Input tensor must be 4D (NCHW), received {len(x.shape)}D instead: {x.shape}")
        input_dtype = x.dtype
        x = x.permute(0, 2, 3, 1)
        x = super().forward(x)
        x = x.permute(0, 3, 1, 2)
        return x.to(input_dtype)


class LayerNorm(nn.LayerNorm):
    def forward(self, x: Tensor) -> Tensor:
        input_dtype = x.dtype
        out = super().forward(x)
        return out.to(input_dtype)






import torch
from torch.distributions import Independent, Normal, Bernoulli
from torch.distributions.utils import logits_to_probs
import numpy as np
from argparse import Namespace
from xuance.common import Any, Dict, Tuple, Sequence, List
from xuance.torch import Tensor, Module
from xuance.torch.utils.distributions import MSEDistribution, SymLogDistribution, TwoHotEncodingDistribution, BernoulliSafeMode
from xuance.torch.utils import dotdict, Moments, compute_lambda_values
from xuance.torch.utils.harmonizer import Harmonizer


class DreamerV3Policy(Module):  # checked
    def __init__(self,
                 model: Module,
                 config: Namespace):
        super(DreamerV3Policy, self).__init__()
        # convert to dotdict
        self.config = dotdict(vars(config))
        self.stoch_size = self.config.world_model.stochastic_size
        self.disc_size = self.config.world_model.discrete_size
        self.stoch_state_size = self.stoch_size * self.disc_size  # 1024 = 32 * 32
        self.batch_size = self.config.batch_size
        self.seq_len = self.config.seq_len
        self.recurrent_state_size = self.config.world_model.recurrent_model.recurrent_state_size
        self.device = self.config.device
        self.is_continuous = self.config.is_continuous
        self.actions_dim = np.sum(self.config.act_shape)  # continuous: num of action props; discrete: num of actions

        # nets
        self.model: Module = model
        self.world_model: Module = self.model.world_model
        self.actor: Module = self.model.actor
        self.critic: Module = self.model.critic
        self.target_critic: Module = self.model.target_critic

        # running mean
        self.moments = Moments(
            self.config.actor.moments.decay,
            self.config.actor.moments.max,
            self.config.actor.moments.percentile.low,
            self.config.actor.moments.percentile.high,
        )

        self.harmonizer_s1 = Harmonizer(self.device)
        self.harmonizer_s2 = Harmonizer(self.device)
        self.harmonizer_s3 = Harmonizer(self.device)

    def model_forward(self,
                      obs: Tensor,
                      acts: Tensor,
                      is_first: Tensor) \
            -> Tuple[SymLogDistribution, TwoHotEncodingDistribution, Independent, Tensor, Tensor,
                     Tensor, Tensor]:
        recurrent_state = torch.zeros(1, self.batch_size, self.recurrent_state_size, device=self.device)  # [1, 16, 512]
        recurrent_states = torch.empty(self.seq_len, self.batch_size, self.recurrent_state_size,
                                       device=self.device)  # [64, 16, 512]
        priors_logits = torch.empty(self.seq_len, self.batch_size, self.stoch_state_size, device=self.device)  # [64, 16, 1024]
        embedded_obs = self.world_model.encoder(obs)  # [64, 16, 512]

        # [1, 16, 32, 32], [64, 16, 32, 32], [64, 16, 1024]
        posterior = torch.zeros(1, self.batch_size, self.stoch_size, self.disc_size, device=self.device)
        posteriors = torch.empty(self.seq_len, self.batch_size, self.stoch_size, self.disc_size, device=self.device)
        posteriors_logits = torch.empty(self.seq_len, self.batch_size, self.stoch_state_size, device=self.device)
        for i in range(0, self.seq_len):
            recurrent_state, posterior, _, posterior_logits, prior_logits = self.world_model.rssm.dynamic(
                posterior,  # z0  [1, 16, 32, 32]
                recurrent_state,  # h0  [1, 16, 512]
                acts[i: i + 1],  # a0  [1, 16, 2]
                embedded_obs[i: i + 1],  # x1  [1, 16, 512]
                is_first[i: i + 1],  # is_first1  [1, 16, 1]
            )  # h0, cat(z0, a0) -> h1; h1 + x1 -> z1; h1 -> z1_hat
            recurrent_states[i] = recurrent_state
            priors_logits[i] = prior_logits  # z1_hat
            posteriors[i] = posterior
            posteriors_logits[i] = posterior_logits  # z1
        latent_states = torch.cat((posteriors.view(*posteriors.shape[:-2], -1), recurrent_states), -1)
        """model_states: [64, 16, 32 * 32 + 512 = 1536]"""
        reconstructed_obs: Tensor = self.world_model.observation_model(latent_states)
        """po(obs, symlog_dist)"""
        po = SymLogDistribution(reconstructed_obs, dims=len(reconstructed_obs.shape[2:]))
        """pr(rews, two_hot_dist)"""
        pr = TwoHotEncodingDistribution(self.world_model.reward_model(latent_states), dims=1)
        """pc(cont, bernoulli_dist)"""
        pc = Independent(BernoulliSafeMode(logits=self.world_model.continue_model(latent_states)), 1)

        # -> [seq, batch, 32, 32]
        priors_logits = priors_logits.view(*priors_logits.shape[:-1], self.stoch_size, self.disc_size)
        posteriors_logits = posteriors_logits.view(*posteriors_logits.shape[:-1], self.stoch_size, self.disc_size)

        return (po, pr, pc, priors_logits, posteriors_logits,
                recurrent_states, posteriors)

    def actor_critic_forward(self,
                             posteriors: Tensor,
                             recurrent_states: Tensor,
                             terms: Tensor) \
            -> Dict[str, List[Any]]:
        imagined_prior = posteriors.detach().reshape(1, -1, self.stoch_state_size)
        recurrent_state = recurrent_states.detach().reshape(1, -1, self.recurrent_state_size)  # [1, 1024, 512]
        imagined_latent_state = torch.cat((imagined_prior, recurrent_state), -1)  # [1, 1024, 1536]
        imagined_trajectories = torch.empty(
            self.config.horizon + 1,
            self.batch_size * self.seq_len,
            self.stoch_state_size + self.recurrent_state_size,
            device=self.device,
        )  # [16, 1024, 1536]
        imagined_trajectories[0] = imagined_latent_state
        imagined_actions = torch.empty(
            self.config.horizon + 1,
            self.batch_size * self.seq_len,
            self.actions_dim,
            device=self.device,
        )  # [16, 1024, 2]
        actions = torch.cat(self.actor(imagined_latent_state.detach())[0], dim=-1)  # z0 -> a0
        imagined_actions[0] = actions

        for i in range(1, self.config.horizon + 1):
            imagined_prior, recurrent_state = self.world_model.rssm.imagination(imagined_prior, recurrent_state, actions)
            imagined_prior = imagined_prior.view(1, -1, self.stoch_state_size)  # [1, 1024, 1024]
            imagined_latent_state = torch.cat((imagined_prior, recurrent_state), -1)
            imagined_trajectories[i] = imagined_latent_state
            actions = torch.cat(self.actor(imagined_latent_state.detach())[0], dim=-1)
            imagined_actions[i] = actions
        predicted_values = TwoHotEncodingDistribution(self.critic(imagined_trajectories), dims=1).mean
        predicted_rewards = TwoHotEncodingDistribution(self.world_model.reward_model(imagined_trajectories), dims=1).mean
        continues = Independent(BernoulliSafeMode(logits=self.world_model.continue_model(imagined_trajectories)), 1).mode
        true_continue = (1 - terms).flatten().reshape(1, -1, 1)  # continues: [16, 1024, 1]; true: [1, 1024, 1]
        continues = torch.cat((true_continue, continues[1:]))
        """seq_shift[1:]"""
        lambda_values = compute_lambda_values(
            predicted_rewards[1:],
            predicted_values[1:],
            continues[1:] * self.config.gamma,
            lmbda=self.config.lmbda,
        )

        with torch.no_grad():
            discount = torch.cumprod(continues * self.config.gamma, dim=0) / self.config.gamma

        policies: Sequence[torch.distributions.Distribution] = self.actor(imagined_trajectories.detach())[1]

        baseline = predicted_values[:-1]
        offset, invscale = self.moments(lambda_values)
        normed_lambda_values = (lambda_values - offset) / invscale
        normed_baseline = (baseline - offset) / invscale
        advantage = normed_lambda_values - normed_baseline
        if self.is_continuous:
            objective = advantage
        else:
            objective = (
                torch.stack(
                    [
                        p.log_prob(imgnd_act.detach()).unsqueeze(-1)[:-1]
                        for p, imgnd_act in zip(policies, torch.split(imagined_actions, [self.actions_dim], dim=-1))
                    ],
                    dim=-1,
                ).sum(dim=-1)
                * advantage.detach()
            )
        try:
            entropy = self.config.actor.ent_coef * torch.stack([p.entropy() for p in policies], -1).sum(dim=-1)
        except NotImplementedError:
            entropy = torch.zeros_like(objective)

        """seq_shift"""
        qv = TwoHotEncodingDistribution(self.critic(imagined_trajectories.detach()[:-1]), dims=1)
        predicted_target_values = TwoHotEncodingDistribution(
            self.target_critic(imagined_trajectories.detach()[:-1]), dims=1
        ).mean
        return {
            'for_actor': [objective, discount, entropy],
            'for_critic': [qv, predicted_target_values, lambda_values]
        }

    def soft_update(self, tau=0.02):  # checked
        for ep, tp in zip(self.critic.parameters(), self.target_critic.parameters()):
            tp.data.mul_(1 - tau)
            tp.data.add_(tau * ep.data)



class DreamerV2Policy(Module):  # checked
    def __init__(self,
                 model: Module,
                 config: Namespace):
        super(DreamerV2Policy, self).__init__()
        # convert to dotdict
        self.config = dotdict(vars(config))
        self.stoch_size = self.config.world_model.stochastic_size
        self.disc_size = self.config.world_model.discrete_size
        self.stoch_state_size = self.stoch_size * self.disc_size  # 1024 = 32 * 32
        self.batch_size = self.config.batch_size
        self.seq_len = self.config.seq_len
        self.recurrent_state_size = self.config.world_model.recurrent_model.recurrent_state_size
        self.device = self.config.device
        self.is_continuous = self.config.is_continuous
        self.actions_dim = np.sum(self.config.act_shape)  # continuous: num of action props; discrete: num of actions

        # nets
        self.model: Module = model
        self.world_model: Module = self.model.world_model
        self.actor: Module = self.model.actor
        self.critic: Module = self.model.critic
        self.target_critic: Module = self.model.target_critic

        self.harmonizer_s1 = Harmonizer(self.device)
        self.harmonizer_s2 = Harmonizer(self.device)
        self.harmonizer_s3 = Harmonizer(self.device)

    def model_forward(self,
                      obs: Tensor,
                      acts: Tensor,
                      is_first: Tensor) \
            -> Tuple[Independent, Independent, Independent, Tensor, Tensor,
                     Tensor, Tensor]:
        recurrent_state = torch.zeros(1, self.batch_size, self.recurrent_state_size, device=self.device)
        recurrent_states = torch.zeros(self.seq_len, self.batch_size, self.recurrent_state_size,
                                       device=self.device)
        priors_logits = torch.empty(self.seq_len, self.batch_size, self.stoch_state_size, device=self.device)
        embedded_obs = self.world_model.encoder(obs)

        posterior = torch.zeros(1, self.batch_size, self.stoch_size, self.disc_size, device=self.device)
        posteriors = torch.empty(self.seq_len, self.batch_size, self.stoch_size, self.disc_size, device=self.device)
        posteriors_logits = torch.empty(self.seq_len, self.batch_size, self.stoch_state_size, device=self.device)
        for i in range(0, self.seq_len):
            recurrent_state, posterior, _, posterior_logits, prior_logits = self.world_model.rssm.dynamic(
                posterior,
                recurrent_state,
                acts[i: i + 1],
                embedded_obs[i: i + 1],
                is_first[i: i + 1],
            )
            recurrent_states[i] = recurrent_state
            priors_logits[i] = prior_logits
            posteriors[i] = posterior
            posteriors_logits[i] = posterior_logits
        latent_states = torch.cat((posteriors.view(*posteriors.shape[:-2], -1), recurrent_states), -1)
        reconstructed_obs: Tensor = self.world_model.observation_model(latent_states)

        po = Independent(Normal(reconstructed_obs, 1), len(reconstructed_obs.shape[2:]))
        pr = Independent(Normal(self.world_model.reward_model(latent_states), 1), 1)
        # error due to not support of Boolean()
        if self.config.world_model.use_continues:
            pc = Independent(Bernoulli(logits=self.world_model.continue_model(latent_states)), 1)
        else:
            pc = None

        # -> [seq, batch, 32, 32]
        priors_logits = priors_logits.view(*priors_logits.shape[:-1], self.stoch_size, self.disc_size)
        posteriors_logits = posteriors_logits.view(*posteriors_logits.shape[:-1], self.stoch_size, self.disc_size)

        return (po, pr, pc, priors_logits, posteriors_logits,
                recurrent_states, posteriors)


    def actor_critic_forward(self,
                             posteriors: Tensor,
                             recurrent_states: Tensor,
                             terms: Tensor) \
            -> Dict[str, List[Any]]:
        imagined_prior = posteriors.detach().reshape(1, -1, self.stoch_state_size)
        recurrent_state = recurrent_states.detach().reshape(1, -1, self.recurrent_state_size)
        imagined_latent_state = torch.cat((imagined_prior, recurrent_state), -1)
        imagined_trajectories = torch.empty(
            self.config.horizon + 1,
            self.batch_size * self.seq_len,
            self.stoch_state_size + self.recurrent_state_size,
            device=self.device,
        )
        imagined_trajectories[0] = imagined_latent_state
        imagined_actions = torch.empty(
            self.config.horizon + 1,
            self.batch_size * self.seq_len,
            self.actions_dim,
            device=self.device,
        )
        # diff to v3; here is, at imagined_trajectories[0] takes action imagined_actions[1]
        imagined_actions[0] = torch.zeros(1, self.batch_size * self.seq_len, self.actions_dim)
        for i in range(1, self.config.horizon + 1):
            # (1, batch_size * seq_len, sum(actions_dim))
            actions = torch.cat(self.actor(imagined_latent_state.detach())[0], dim=-1)
            imagined_actions[i] = actions
            imagined_prior, recurrent_state = self.world_model.rssm.imagination(imagined_prior, recurrent_state, actions)
            imagined_prior = imagined_prior.view(1, -1, self.stoch_state_size)
            imagined_latent_state = torch.cat((imagined_prior, recurrent_state), -1)
            imagined_trajectories[i] = imagined_latent_state
        predicted_target_values = self.target_critic(imagined_trajectories)
        predicted_rewards = self.world_model.reward_model(imagined_trajectories)
        if self.config.world_model.use_continues:
            continues = logits_to_probs(self.world_model.continue_model(imagined_trajectories), is_binary=True)  # diff to v3
            # diff to v3(v3: no self.config.gamma here, but mult gamma before passing to 'compute_lambda_values')
            true_continue = (1 - terms).reshape(1, -1, 1) * self.config.gamma
            continues = torch.cat((true_continue, continues[1:]))
        else:
            continues = torch.ones_like(predicted_rewards.detach()) * self.config.gamma

        # Compute the lambda_values, by passing as last value the value of the last imagined state
        # (horizon, batch_size * seq_len, 1)
        lambda_values = DreamerV2Policy.compute_lambda_values(
            predicted_rewards[:-1],
            predicted_target_values[:-1],
            continues[:-1],
            bootstrap=predicted_target_values[-1:],
            horizon=self.config.horizon,
            lmbda=self.config.lmbda,
        )

        with torch.no_grad():
            discount = torch.cumprod(torch.cat((torch.ones_like(continues[:1]), continues[:-1]), 0), 0)

        policies: Sequence[torch.distributions.Distribution] = self.actor(imagined_trajectories[:-2].detach())[1]
        # Dynamics backpropagation
        dynamics = lambda_values[1:]
        # Reinforce
        advantage = (lambda_values[1:] - predicted_target_values[:-2]).detach()
        reinforce = (
                torch.stack(
                    [
                        p.log_prob(imgnd_act[1:-1].detach()).unsqueeze(-1)
                        for p, imgnd_act in zip(policies, torch.split(imagined_actions, [self.actions_dim], -1))
                    ],
                    -1,
                ).sum(-1)
                * advantage
        )
        objective = self.config.actor.objective_mix * reinforce + (1 - self.config.actor.objective_mix) * dynamics
        try:
            entropy = self.config.actor.ent_coef * torch.stack([p.entropy() for p in policies], -1).sum(dim=-1)
        except NotImplementedError:
            entropy = torch.zeros_like(objective)
        # policy_loss = -torch.mean(discount[:-2].detach() * (objective + entropy.unsqueeze(-1)))

        # last imagined state (with position=horizon+1) in the trajectory only used for bootstrapping;
        qv = Independent(Normal(self.critic(imagined_trajectories.detach()[:-1]), 1), 1)

        return {
            'for_actor': [objective, discount, entropy],
            'for_critic': [qv, predicted_target_values, lambda_values]
        }

    def hard_update(self):  # checked
        for ep, tp in zip(self.critic.parameters(), self.target_critic.parameters()):
            tp.data.mul_(0)
            tp.data.add_(1.0 * ep.data)

    @staticmethod
    def compute_lambda_values(
            rewards: Tensor,
            values: Tensor,
            continues: Tensor,
            bootstrap: Tensor = None,
            horizon: int = 15,
            lmbda: float = 0.95,
    ) -> Tensor:
        if bootstrap is None:
            bootstrap = torch.zeros_like(values[-1:])
        agg = bootstrap
        next_val = torch.cat((values[1:], bootstrap), dim=0)
        inputs = rewards + continues * next_val * (1 - lmbda)
        lv = []
        for i in reversed(range(horizon)):
            agg = inputs[i] + continues[i] * lmbda * agg
            lv.append(agg)
        return torch.cat(list(reversed(lv)), dim=0)


"""
Soft Actor-Critic with discrete action spaces (SAC-Discrete)
Paper link: https://arxiv.org/pdf/1910.07207.pdf
Implementation: TensorFlow2
"""
import numpy as np
from argparse import Namespace
from xuance.common import Optional
from xuance.tensorflow import tf, tk, Module
from xuance.tensorflow.learners import Learner


class AlphaLayer(Module):
    def __init__(self, action_dim):
        super(AlphaLayer, self).__init__()
        self.log_alpha = self.add_weight(name="log_of_alpha", shape=(action_dim,), initializer=tf.zeros, trainable=True)


class SACDIS_Learner(Learner):
    def __init__(self,
                 config: Namespace,
                 policy: Module,
                 target_entropy: Optional[float] = None):
        super(SACDIS_Learner, self).__init__(config, policy)
        if ("macOS" in self.os_name) and ("arm" in self.os_name):  # For macOS with Apple's M-series chips.
            if self.distributed_training:
                with self.policy.mirrored_strategy.scope():
                    self.optimizer = {'actor': tk.optimizers.legacy.Adam(config.learning_rate_actor),
                                      'critic': tk.optimizers.legacy.Adam(config.learning_rate_critic)}
            else:
                self.optimizer = {'actor': tk.optimizers.legacy.Adam(config.learning_rate_actor),
                                  'critic': tk.optimizers.legacy.Adam(config.learning_rate_critic)}
        else:
            if self.distributed_training:
                with self.policy.mirrored_strategy.scope():
                    self.optimizer = {'actor': tk.optimizers.Adam(config.learning_rate_actor),
                                      'critic': tk.optimizers.Adam(config.learning_rate_critic)}
            else:
                self.optimizer = {'actor': tk.optimizers.Adam(config.learning_rate_actor),
                                  'critic': tk.optimizers.Adam(config.learning_rate_critic)}
        self.tau = config.tau
        self.gamma = config.gamma
        self.alpha = config.alpha
        self.use_automatic_entropy_tuning = config.use_automatic_entropy_tuning
        if self.use_automatic_entropy_tuning:
            self.target_entropy = target_entropy
            if self.distributed_training:
                with self.policy.mirrored_strategy.scope():
                    self.alpha_layer = AlphaLayer(1)
                    self.alpha = tf.exp(self.alpha_layer.log_alpha)
                    if ("macOS" in self.os_name) and ("arm" in self.os_name):  # For macOS with Apple's M-series chips.
                        self.alpha_optimizer = tk.optimizers.legacy.Adam(config.learning_rate_actor)
                    else:
                        self.alpha_optimizer = tk.optimizers.Adam(config.learning_rate_actor)
            else:
                self.alpha_layer = AlphaLayer(1)
                self.alpha = tf.exp(self.alpha_layer.log_alpha)
                if ("macOS" in self.os_name) and ("arm" in self.os_name):  # For macOS with Apple's M-series chips.
                    self.alpha_optimizer = tk.optimizers.legacy.Adam(config.learning_rate_actor)
                else:
                    self.alpha_optimizer = tk.optimizers.Adam(config.learning_rate_actor)

    @tf.function
    def actor_forward_fn(self, obs_batch):
        with tf.GradientTape() as tape:
            action_prob, log_pi, policy_q_1, policy_q_2 = self.policy.Qpolicy(obs_batch)
            policy_q = tf.math.minimum(policy_q_1, policy_q_2)
            p_loss = tf.reduce_mean(tf.reduce_sum(action_prob * (self.alpha * log_pi - policy_q), axis=-1))
            gradients = tape.gradient(p_loss, self.policy.actor_trainable_variables)
            if self.use_grad_clip:
                self.optimizer['actor'].apply_gradients([
                    (tf.clip_by_norm(grad, self.grad_clip_norm), var)
                    for (grad, var) in zip(
                        gradients, self.policy.actor_trainable_variables)
                    if grad is not None])
            else:
                self.optimizer['actor'].apply_gradients([
                    (grad, var)
                    for (grad, var) in zip(
                        gradients, self.policy.actor_trainable_variables)
                    if grad is not None])
        return p_loss, log_pi, policy_q

    @tf.function
    def critic_forward_fn(self, obs_batch, act_batch, rew_batch, next_batch, ter_batch):
        with tf.GradientTape() as tape:
            action_q_1, action_q_2 = self.policy.Qaction(obs_batch)
            action_q_1 = tf.gather(params=action_q_1, indices=act_batch, axis=-1, batch_dims=-1)
            action_q_2 = tf.gather(params=action_q_2, indices=act_batch, axis=-1, batch_dims=-1)
            action_prob_next, log_pi_next, target_q = self.policy.Qtarget(next_batch)
            target_q = action_prob_next * (target_q - self.alpha * log_pi_next)
            target_q = tf.expand_dims(tf.reduce_sum(target_q, axis=1), axis=-1)
            rew = tf.expand_dims(rew_batch, axis=-1)
            backup = rew + (1 - ter_batch) * self.gamma * target_q
            y_true = tf.stop_gradient(tf.reshape(backup, [-1]))
            y_pred_1 = tf.reshape(action_q_1, [-1])
            y_pred_2 = tf.reshape(action_q_2, [-1])
            q_loss = tk.losses.mean_squared_error(y_true, y_pred_1) + tk.losses.mean_squared_error(y_true, y_pred_2)
            gradients = tape.gradient(q_loss, self.policy.critic_trainable_variables)
            if self.use_grad_clip:
                self.optimizer['critic'].apply_gradients([
                    (tf.clip_by_norm(grad, self.grad_clip_norm), var)
                    for (grad, var) in zip(gradients, self.policy.critic_trainable_variables)
                    if grad is not None])
            else:
                self.optimizer['critic'].apply_gradients([
                    (grad, var)
                    for (grad, var) in zip(gradients, self.policy.critic_trainable_variables)
                    if grad is not None])
        return q_loss

    @tf.function
    def alpha_forward_fn(self, log_pi):
        with tf.GradientTape() as tape:
            alpha_loss = -tf.math.reduce_mean(self.alpha_layer.log_alpha.value() * (log_pi + self.target_entropy))
            gradients = tape.gradient(alpha_loss, self.alpha_layer.trainable_variables)
            self.alpha_optimizer.apply_gradients([
                (grad, var)
                for (grad, var) in zip(gradients, self.alpha_layer.trainable_variables)
                if grad is not None
            ])
        return alpha_loss

    @tf.function
    def learn_actor(self, *inputs):
        if self.distributed_training:
            p_loss, log_pi, policy_q = self.policy.mirrored_strategy.run(self.actor_forward_fn, args=inputs)
            return (self.policy.mirrored_strategy.reduce(tf.distribute.ReduceOp.SUM, p_loss, axis=None),
                    self.policy.mirrored_strategy.reduce(tf.distribute.ReduceOp.SUM, log_pi, axis=None),
                    self.policy.mirrored_strategy.reduce(tf.distribute.ReduceOp.SUM, policy_q, axis=None))
        else:
            return self.actor_forward_fn(*inputs)

    @tf.function
    def learn_critic(self, *inputs):
        if self.distributed_training:
            q_loss = self.policy.mirrored_strategy.run(self.critic_forward_fn, args=inputs)
            return self.policy.mirrored_strategy.reduce(tf.distribute.ReduceOp.SUM, q_loss, axis=None)
        else:
            return self.critic_forward_fn(*inputs)

    @tf.function
    def learn_alpha(self, *inputs):
        if self.distributed_training:
            alpha_loss = self.policy.mirrored_strategy.run(self.alpha_forward_fn, args=inputs)
            return self.policy.mirrored_strategy.reduce(tf.distribute.ReduceOp.SUM, alpha_loss, axis=None)
        else:
            return self.alpha_forward_fn(*inputs)

    def update(self, **samples):
        self.iterations += 1
        obs_batch = samples['obs']
        act_batch = samples['actions'].astype(np.int32)
        next_batch = samples['obs_next']
        rew_batch = samples['rewards']
        ter_batch = samples['terminals']

        q_loss = self.learn_critic(obs_batch, act_batch, rew_batch, next_batch, ter_batch)
        p_loss, log_pi, policy_q = self.learn_actor(obs_batch)
        if self.use_automatic_entropy_tuning:
            alpha_loss = self.learn_alpha(log_pi)
            alpha_loss = alpha_loss.numpy()
            self.alpha = tf.math.exp(self.alpha_layer.log_alpha).numpy()
        else:
            alpha_loss = 0

        self.policy.soft_update(self.tau)

        info = {
            "Qloss": q_loss.numpy(),
            "Ploss": p_loss.numpy(),
            "Qvalue": tf.reduce_mean(policy_q).numpy(),
            "alpha_loss": alpha_loss,
            "alpha": self.alpha,
        }

        return info



from tqdm import tqdm
from copy import deepcopy
from argparse import Namespace
from xuance.common import Union, Optional, PerOffPolicyBuffer
from xuance.environment import DummyVecEnv, SubprocVecEnv
from xuance.torch.agents import BaseCallback
from xuance.torch.agents.qlearning_family import DQN_Agent


class PerDQN_Agent(DQN_Agent):
    """The implementation of Per-DQN agent.

    Args:
        config: the Namespace variable that provides hyperparameters and other settings.
        envs: the vectorized environments.
        callback: A user-defined callback function object to inject custom logic during training.
    """

    def __init__(self,
                 config: Namespace,
                 envs: Union[DummyVecEnv, SubprocVecEnv],
                 callback: Optional[BaseCallback] = None):
        super(PerDQN_Agent, self).__init__(config, envs, callback)
        self.PER_beta0 = config.PER_beta0
        self.PER_beta = config.PER_beta0

        # Create experience replay buffer.
        self.auxiliary_info_shape = {}
        self.atari = True if config.env_name == "Atari" else False
        self.memory = PerOffPolicyBuffer(observation_space=self.observation_space,
                                         action_space=self.action_space,
                                         auxiliary_shape=self.auxiliary_info_shape,
                                         n_envs=self.n_envs,
                                         buffer_size=config.buffer_size,
                                         batch_size=config.batch_size,
                                         alpha=config.PER_alpha)
        self.learner = self._build_learner(self.config, self.policy, self.callback)

    def train_epochs(self, n_epochs=1):
        train_info = {}
        for _ in range(n_epochs):
            samples = self.memory.sample(self.PER_beta)
            td_error, step_info = self.learner.update(**samples)
            self.memory.update_priorities(samples['step_choices'], td_error)
        train_info["epsilon-greedy"] = self.e_greedy
        return train_info

    def train(self, train_steps):
        train_info = {}
        obs = self.envs.buf_obs
        for _ in tqdm(range(train_steps)):
            self.obs_rms.update(obs)
            obs = self._process_observation(obs)
            policy_out = self.action(obs, test_mode=False)
            acts = policy_out['actions']
            next_obs, rewards, terminals, truncations, infos = self.envs.step(acts)

            self.callback.on_train_step(self.current_step, envs=self.envs, policy=self.policy,
                                        obs=obs, policy_out=policy_out, acts=acts, next_obs=next_obs, rewards=rewards,
                                        terminals=terminals, truncations=truncations, infos=infos,
                                        train_steps=train_steps)

            self.memory.store(obs, acts, self._process_reward(rewards), terminals, self._process_observation(next_obs))
            if self.current_step > self.start_training and self.current_step % self.training_frequency == 0:
                update_info = self.train_epochs(n_epochs=self.n_epochs)
                self.log_infos(update_info, self.current_step)
                train_info.update(update_info)
                self.PER_beta += (1 - self.PER_beta0) / train_steps
                self.callback.on_train_epochs_end(self.current_step, policy=self.policy, memory=self.memory,
                                                  current_episode=self.current_episode, train_steps=train_steps,
                                                  update_info=update_info, per_beta=self.PER_beta)

            obs = deepcopy(next_obs)
            for i in range(self.n_envs):
                if terminals[i] or truncations[i]:
                    if self.atari and (~truncations[i]):
                        pass
                    else:
                        obs[i] = infos[i]["reset_obs"]
                        self.envs.buf_obs[i] = obs[i]
                        self.current_episode[i] += 1
                        if self.use_wandb:
                            episode_info = {
                                f"Episode-Steps/rank_{self.rank}/env-{i}": infos[i]["episode_step"],
                                f"Train-Episode-Rewards/rank_{self.rank}/env-{i}": infos[i]["episode_score"]
                            }
                        else:
                            episode_info = {
                                f"Episode-Steps/rank_{self.rank}": {f"env-{i}": infos[i]["episode_step"]},
                                f"Train-Episode-Rewards/rank_{self.rank}": {f"env-{i}": infos[i]["episode_score"]}
                            }
                        self.log_infos(episode_info, self.current_step)
                        train_info.update(episode_info)
                        self.callback.on_train_episode_info(envs=self.envs, policy=self.policy, env_id=i,
                                                            infos=infos, rank=self.rank, use_wandb=self.use_wandb,
                                                            current_step=self.current_step,
                                                            current_episode=self.current_episode,
                                                            train_steps=train_steps)

            self.current_step += self.n_envs
            if self.e_greedy > self.end_greedy:
                self.e_greedy -= self.delta_egreedy
            self.callback.on_train_step_end(self.current_step, envs=self.envs, policy=self.policy,
                                            train_steps=train_steps, train_info=train_info)
        return train_info



from torch import optim

import torch
import torch.nn as nn
from xuance.torch.learners import Learner
import torch.nn.functional as F
try:
    from torchvision import transforms
except:
    pass


class FrameStackTransform:
    def __init__(self):
        self.transform = transforms.Compose([
            transforms.RandomApply([transforms.ColorJitter(0.4, 0.4, 0.4)], p=0.8),
            transforms.RandomGrayscale(p=0.2),
            transforms.RandomHorizontalFlip(),
            transforms.RandomAdjustSharpness(2),
            transforms.RandomPerspective(0.2),
            transforms.RandomRotation(15),
        ])

    def __call__(self, x: torch.Tensor):
        # x: (B, H, W, C), C=4
        x = x.permute(0,3,1,2)
        B, C, H, W = x.shape
        x_merged = x.contiguous().reshape(B*C, 1, H, W)  # (B*C,1,H,W)
        x_transformed = self.transform(x_merged)
        _, _, new_H, new_W = x_transformed.shape
        x_transformed = x_transformed.view(B, C, new_H, new_W)
        return x_transformed.permute(0,2,3,1)

class SPR_Learner(Learner):

    def __init__(self, config, policy, callback, temperature=0.1, tau=0.99, repr_lr=1e-4, prediction_steps=3):
        super().__init__(config, policy, callback)
        self.temperature = temperature
        self.tau = tau
        self.prediction_steps = prediction_steps

        self.encoder_optim = optim.Adam(policy.representation.parameters(), lr=repr_lr)
        self.q_optim = optim.Adam(policy.parameters(), lr=config.learning_rate)
        self.transition_model = nn.GRUCell(64 + self.policy.action_dim, 64).to(config.device)
        self.mse_loss = nn.MSELoss()
        self.transform = FrameStackTransform()

    def _update_target_encoder(self):
        for target_param, param in zip(self.policy.target_representation.parameters(),
                                       self.policy.representation.parameters()):
            target_param.data.copy_(self.tau * target_param.data + (1 - self.tau) * param.data)

    def _compute_contrastive_loss(self, obs: torch.Tensor, actions: torch.Tensor):
        aug_obs = self.transform(obs)
        q = self.policy.representation(aug_obs)['state']
        with torch.no_grad():
            k = self.policy.target_representation(obs)['state']
            k = nn.functional.normalize(k, dim=1)

        predicted_latents = [q]
        hx = q.clone()
        actions_one_hot = F.one_hot(actions.long(), num_classes=self.policy.action_dim).float().to(q.device)
        for _ in range(self.prediction_steps):
            hx = self.transition_model(torch.cat([hx, actions_one_hot], dim=1), hx)
            predicted_latents.append(hx)

        total_loss = 0
        for t in range(1, self.prediction_steps + 1):
            logits = torch.mm(predicted_latents[t], k.T) / self.temperature
            labels = torch.arange(obs.size(0), device=obs.device)
            total_loss += nn.CrossEntropyLoss()(logits, labels)
        return total_loss / self.prediction_steps

    def update(self, **samples):
        obs = torch.as_tensor(samples['obs'], device=self.device)
        actions = torch.as_tensor(samples['actions'], device=self.device)
        next_obs = torch.as_tensor(samples['obs_next'], device=self.device)
        rew = torch.as_tensor(samples['rewards'], device=self.device)
        done = torch.as_tensor(samples['terminals'], dtype=torch.float, device=self.device)
        info = self.callback.on_update_start(self.iterations,
                                             policy=self.policy, obs=obs, act=actions,
                                             next_obs=next_obs, rew=rew, termination=done)

        spr_loss = self._compute_contrastive_loss(obs, actions)
        self.encoder_optim.zero_grad()
        spr_loss.backward()
        if self.use_grad_clip:
            torch.nn.utils.clip_grad_norm_(self.policy.representation.parameters(), self.grad_clip_norm)
        self.encoder_optim.step()
        self._update_target_encoder()

        _, _, evalQ = self.policy(obs)
        _, _, targetQ = self.policy.target(next_obs)
        predictQ = rew + self.gamma * (1 - done) * targetQ.max(dim=1).values
        q_loss = self.mse_loss(evalQ.gather(1, actions.long().unsqueeze(1)).squeeze(), predictQ)

        self.q_optim.zero_grad()
        q_loss.backward()
        if self.use_grad_clip:
            torch.nn.utils.clip_grad_norm_(self.policy.parameters(), self.grad_clip_norm)
        self.q_optim.step()

        # print("spr_loss:", spr_loss.item(), "q_loss:", q_loss.item())
        info.update({
            "spr_loss": spr_loss.item(),
            "q_loss": q_loss.item(),
            "learning_rate": self.q_optim.param_groups[0]['lr']
        })

        info.update(self.callback.on_update_end(self.iterations,
                                                policy=self.policy, info=info,
                                                spr_loss=spr_loss, q_loss=q_loss,
                                                evalQ=evalQ, predictQ=predictQ, targetQ=targetQ))

        return info



"""
Proximal Policy Optimization with KL divergence (PPO-KL)
Paper link: https://arxiv.org/pdf/1707.06347.pdf
Implementation: MindSpore
"""
from xuance.mindspore import ms, Module, Tensor, optim
from xuance.mindspore.learners import Learner
from xuance.mindspore.utils import clip_grads
from xuance.mindspore.utils.operations import merge_distributions
from argparse import Namespace
from mindspore.nn import MSELoss


class PPOKL_Learner(Learner):
    def __init__(self,
                 config: Namespace,
                 policy: Module):
        super(PPOKL_Learner, self).__init__(config, policy)
        self.optimizer = optim.Adam(params=self.policy.trainable_params(), lr=self.config.learning_rate, eps=1e-5)
        self.scheduler = optim.lr_scheduler.LinearLR(self.optimizer, start_factor=1.0, end_factor=self.end_factor_lr_decay,
                                                     total_iters=self.config.running_steps)
        self.vf_coef = config.vf_coef
        self.ent_coef = config.ent_coef
        self.target_kl = config.target_kl
        self.kl_coef = config.kl_coef
        self.clip_range = config.clip_range

        self._clip_range = [Tensor(1.0 - self.clip_range), Tensor(1.0 + self.clip_range)]
        self._exp = ms.ops.Exp()
        self._minimum = ms.ops.Minimum()
        self._mean = ms.ops.ReduceMean(keep_dims=True)
        self._loss = MSELoss()
        # Get gradient function
        self.grad_fn = ms.value_and_grad(self.forward_fn, None, self.optimizer.parameters, has_aux=True)
        self.policy.set_train()
        self.continuous_control = True if ("gaussian" in str(type(self.policy))) else False

    def forward_fn(self, obs_batch, act_batch, ret_batch, adv_batch, old_dists_param):
        old_log_prob = old_dists_param['log_prob']
        outputs, act_dist, v_pred = self.policy(obs_batch)
        log_prob = act_dist.log_prob(act_batch)
        ratio = self._exp(log_prob - old_log_prob)
        if self.continuous_control:
            mean, std = old_dists_param['mean'], old_dists_param['std']
            kl = act_dist.distribution.kl_loss('Normal', mean, std).mean()
        else:
            probs = old_dists_param['probs']
            kl = act_dist.distribution.kl_loss('Categorical', probs).mean()
        loss_a = -(ratio * adv_batch).mean() + self.kl_coef * kl
        loss_c = self._loss(logits=v_pred, labels=ret_batch)
        loss_e = self._mean(act_dist.entropy())
        loss = loss_a - self.ent_coef * loss_e + self.vf_coef * loss_c
        return loss, loss_a, loss_c, loss_e, kl, v_pred

    def update(self, **samples):
        self.iterations += 1
        obs_batch = Tensor(samples['obs'])
        act_batch = Tensor(samples['actions'])
        ret_batch = Tensor(samples['returns'])
        adv_batch = Tensor(samples['advantages'])
        old_dist = merge_distributions(samples['aux_batch']['old_dist'])
        old_logp_batch = old_dist.log_prob(act_batch)
        if self.continuous_control:
            old_dists_param = {
                "mean": old_dist.mu,
                "std": old_dist.std,
                "log_prob": old_logp_batch
            }
        else:
            old_dists_param = {
                "probs": old_dist.probs,
                "log_prob": old_logp_batch
            }

        (loss, loss_a, loss_c, loss_e, kl, v_pred), grads = self.grad_fn(obs_batch, act_batch, ret_batch, adv_batch,
                                                                         old_dists_param)
        if self.use_grad_clip:
            grads = clip_grads(grads, Tensor(-self.grad_clip_norm), Tensor(self.grad_clip_norm))
        self.optimizer(grads)

        self.scheduler.step()
        lr = self.scheduler.get_last_lr()[0]

        info = {
            "actor-loss": loss_a.asnumpy(),
            "critic-loss": loss_c.asnumpy(),
            "entropy": loss_e.asnumpy(),
            "learning_rate": lr.asnumpy(),
            "kl": kl.asnumpy(),
            "predict_value": v_pred.mean().asnumpy()
        }

        return info



import gymnasium as gym
import numpy as np
from collections import deque
from xuance.environment.single_agent_env.gym import LazyFrames
try:
    import cv2
except ImportError:
    pass

try:
    import ale_py
    gym.register_envs(ale_py)
except ImportError:
    pass


class Atari_Env(gym.Wrapper):
    """Modified Atari environment with training optimizations.

    Implements several enhancements from DeepMind's Nature paper:
    - Episode termination: Treats end-of-life as end-of-episode (resets only on true game over)
    - Frame skipping: Returns only every `skip`-th frame
    - Observation resizing: Warps frames from 210x160 to configurable size
    - Frame stacking: Efficiently stacks last k frames using lazy arrays

    Note:
        All configurations should be set in the provided config object.
    """
    def __init__(self, config, **kwargs):
        """Initializes the Atari environment wrapper.

        Args:
            config: Configuration object containing:
                env_id: Atari environment ID (e.g., "ALE/Breakout-v5")
                env_seed: Random seed for environment
                obs_type: Observation type ("ram"/"rgb"/"grayscale")
                frame_skip: Frame skip interval (int or tuple for stochastic skipping)
                num_stack: Number of frames to stack
                img_size: Target observation dimensions (default: [210, 160])
                noop_max: Maximum no-op actions during reset
                render_mode: Rendering mode (None/"human"/"rgb_array")
                full_action_space: Whether to use full action space
            **kwargs: Additional arguments passed to gym.make()
        """
        full_action_space = config.full_action_space if hasattr(config, 'full_action_space') else False
        self.env = gym.make(config.env_id,
                            render_mode=config.render_mode,
                            obs_type=config.obs_type,
                            frameskip=config.frame_skip,
                            full_action_space=full_action_space)
        self.env.metadata['render_fps'] = config.fps
        self.env.action_space.seed(seed=config.env_seed)
        self.env.reset(seed=config.env_seed)
        self.max_episode_steps = self.env._max_episode_steps if hasattr(self.env, '_max_episode_steps') else 1e5
        super(Atari_Env, self).__init__(self.env)
        # self.env.seed(seed)
        self.num_stack = config.num_stack
        self.obs_type = config.obs_type
        self.frames = deque([], maxlen=self.num_stack)
        self.image_size = [210, 160] if config.img_size is None else config.img_size
        self.noop_max = config.noop_max
        self.lifes = self.env.unwrapped.ale.lives()
        self.was_real_done = True
        self.grayscale, self.rgb = False, False
        if self.obs_type == "rgb":
            self.rgb = True
            self.observation_space = gym.spaces.Box(low=0, high=255,
                                                    shape=(config.img_size[0], config.img_size[1], 3 * self.num_stack),
                                                    dtype=np.uint8)
        elif self.obs_type == "grayscale":
            self.grayscale = True
            self.observation_space = gym.spaces.Box(low=0, high=255,
                                                    shape=(config.img_size[0], config.img_size[1], self.num_stack),
                                                    dtype=np.uint8)
        else:  # ram type
            self.observation_space = self.env.observation_space
        # assert self.env.unwrapped.get_action_meanings()[0] == "NOOP"
        # assert self.env.unwrapped.get_action_meanings()[1] == "FIRE"
        # assert len(self.env.unwrapped.get_action_meanings()) >= 3
        self.action_space = self.env.action_space
        self.metadata = self.env.metadata
        self._render_mode = config.render_mode
        self._episode_step = 0
        self._episode_score = 0.0

    def close(self):
        """Closes the underlying environment and releases resources."""
        self.env.close()

    def render(self, *args, **kwargs):
        """Renders the environment.

        Returns:
            Rendered frame according to specified mode
        """
        return self.env.render()

    def reset(self, *args):
        """Resets the environment with random no-op actions.

        Performs:
        1. Environment reset
        2. Random number of no-op actions
        3. Initial fire action (if available)
        4. Frame stacking initialization

        Returns:
            tuple: (stacked_observations, info_dict)
        """
        info = {}
        if self.was_real_done:
            self.env.reset()
            # Execute NoOp actions
            num_noops = np.random.randint(0, self.noop_max)
            for _ in range(num_noops):
                obs, _, done, _, _ = self.env.step(0)
                if done:
                    self.env.reset()
            # try to fire
            obs, _, done, _, _ = self.env.step(1)
            if done:
                obs = self.env.reset()
            # stack reset observations
            for _ in range(self.num_stack):
                self.frames.append(self.observation(obs))

            self._episode_step = 0
            self._episode_score = 0.0
            info["episode_step"] = 0
        else:
            obs, _, done, _, _ = self.env.step(0)
            for _ in range(self.num_stack):
                self.frames.append(self.observation(obs))

        self.lifes = self.env.unwrapped.ale.lives()
        self.was_real_done = False
        return self._get_obs(), info

    def step(self, actions):
        """Executes one environment step.

        Args:
            actions: Action to execute

        Returns:
            tuple: (observation, reward, terminated, truncated, info)
        """
        observation, reward, terminated, truncated, info = self.env.step(actions)
        self.frames.append(self.observation(observation))
        lives = self.env.unwrapped.ale.lives()
        # avoid environment bug
        if self._episode_step >= self.max_episode_steps:
            terminated = True
        self.was_real_done = terminated
        if (lives < self.lifes) and (lives > 0):
            terminated = True
        truncated = self.was_real_done
        self.lifes = lives
        self._episode_step += 1
        self._episode_score += reward
        info["episode_score"] = self._episode_score
        info["episode_step"] = self._episode_step
        return self._get_obs(), self.reward(reward), terminated, truncated, info

    def _get_obs(self):
        """Returns stacked observations as LazyFrames.

        Returns:
            LazyFrames: Stacked frame observations

        Raises:
            AssertionError: If frame stack is incomplete
        """
        assert len(self.frames) == self.num_stack
        return LazyFrames(list(self.frames))

    def observation(self, frame):
        """Processes raw frame into desired observation format.

        Args:
            frame: Raw environment frame

        Returns:
            Processed observation (resized grayscale/RGB or raw RAM)
        """
        if self.grayscale:
            return np.expand_dims(cv2.resize(frame, self.image_size, interpolation=cv2.INTER_AREA), -1)
        elif self.rgb:
            return cv2.resize(frame, self.image_size, interpolation=cv2.INTER_AREA)
        else:
            return frame

    def reward(self, reward):
        """Applies reward shaping for training.

        Args:
            reward: Original environment reward

        Returns:
            Shaped reward (sign function in this implementation)
        """
        return np.sign(reward)



import gymnasium as gym
import torch
import numpy as np
from tqdm import tqdm
from copy import deepcopy
from argparse import Namespace
from gymnasium import spaces
from xuance.common import Optional
from xuance.environment.single_agent_env import Gym_Env
from xuance.torch import Module
from xuance.torch.utils import NormalizeFunctions, ActivationFunctions
from xuance.torch.policies import REGISTRY_Policy
from xuance.torch.agents import Agent, BaseCallback
from xuance.torch.agents.policy_gradient.pdqn_agent import PDQN_Agent
from xuance.common import DummyOffPolicyBuffer


class SPDQN_Agent(PDQN_Agent, Agent):
    """The implementation of SPDQN agent.

    Args:
        config: the Namespace variable that provides hyperparameters and other settings.
        envs: the vectorized environments.
        callback: A user-defined callback function object to inject custom logic during training.
    """
    def __init__(self,
                 config: Namespace,
                 envs: Gym_Env,
                 callback: Optional[BaseCallback] = None):
        Agent.__init__(self, config, envs, callback)
        self.start_noise, self.end_noise = config.start_noise, config.end_noise
        self.noise_scale = config.start_noise
        self.delta_noise = (self.start_noise - self.end_noise) / (config.running_steps / self.n_envs)

        self.observation_space = envs.observation_space.spaces[0]
        old_as = envs.action_space
        num_disact = old_as.spaces[0].n
        self.action_space = gym.spaces.Tuple((old_as.spaces[0],
                                              *(gym.spaces.Box(old_as.spaces[1].spaces[i].low,
                                                               old_as.spaces[1].spaces[i].high, dtype=np.float32)
                                                for i in range(0, num_disact))))
        self.action_high = [self.action_space.spaces[i].high for i in range(1, num_disact + 1)]
        self.action_low = [self.action_space.spaces[i].low for i in range(1, num_disact + 1)]
        self.action_range = [self.action_space.spaces[i].high - self.action_space.spaces[i].low for i in
                             range(1, num_disact + 1)]
        self.representation_info_shape = {'state': (envs.observation_space.spaces[0].shape)}
        self.auxiliary_info_shape = {}
        self.nenvs = 1
        self.epsilon = 1.0
        self.epsilon_steps = 1000
        self.epsilon_initial = 1.0
        self.epsilon_final = 0.1
        self.buffer_action_space = spaces.Box(np.zeros(4), np.ones(4), dtype=np.float64)

        # Build policy, optimizer, scheduler.
        self.policy = self._build_policy()

        self.memory = DummyOffPolicyBuffer(observation_space=self.observation_space,
                                           action_space=self.buffer_action_space,
                                           auxiliary_shape=self.auxiliary_info_shape,
                                           n_envs=self.n_envs,
                                           buffer_size=config.buffer_size,
                                           batch_size=config.batch_size)
        self.learner = self._build_learner(self.config, self.policy, self.callback)

        self.num_disact = self.action_space.spaces[0].n
        self.conact_sizes = np.array([self.action_space.spaces[i].shape[0] for i in range(1, self.num_disact + 1)])
        self.conact_size = int(self.conact_sizes.sum())

    def _build_policy(self) -> Module:
        normalize_fn = NormalizeFunctions[self.config.normalize] if hasattr(self.config, "normalize") else None
        initializer = torch.nn.init.orthogonal_
        activation = ActivationFunctions[self.config.activation]
        device = self.device

        # build representation.
        representation = self._build_representation(self.config.representation, self.observation_space, self.config)

        # build policy.
        if self.config.policy == "SPDQN_Policy":
            policy = REGISTRY_Policy["SPDQN_Policy"](
                observation_space=self.observation_space, action_space=self.action_space,
                representation=representation,
                conactor_hidden_size=self.config.conactor_hidden_size,
                qnetwork_hidden_size=self.config.qnetwork_hidden_size,
                normalize=normalize_fn, initialize=initializer, activation=activation, device=device,
                use_distributed_training=self.distributed_training,
                activation_action=ActivationFunctions[self.config.activation_action])
        else:
            raise AttributeError(
                f"{self.config.agent} currently does not support the policy named {self.config.policy}.")

        return policy

    def train(self, train_steps=10000):
        train_info = {}
        episodes = np.zeros((self.nenvs,), np.int32)
        scores = np.zeros((self.nenvs,), np.float32)
        obs, _ = self.envs.reset()
        for _ in tqdm(range(train_steps)):
            step_info = {}
            disaction, conaction, con_actions = self.action(obs)
            action = self.pad_action(disaction, conaction)
            action[1][disaction] = self.action_range[disaction] * (action[1][disaction] + 1) / 2. + self.action_low[
                disaction]
            (next_obs, steps), rewards, terminal, _ = self.envs.step(action)
            if self.render: self.envs.render("human")
            acts = np.concatenate(([disaction], con_actions), axis=0).ravel()

            self.callback.on_train_step(self.current_step, envs=self.envs, policy=self.policy,
                                        obs=obs, next_obs=next_obs, rewards=rewards, terminals=terminal,
                                        action=action, acts=acts, steps=steps,
                                        disaction=disaction, conaction=conaction, con_actions=con_actions,
                                        train_steps=train_steps)

            self.memory.store(obs, acts, rewards, terminal, next_obs)
            if self.current_step > self.start_training and self.current_step % self.training_frequency == 0:
                update_info = self.train_epochs(n_epochs=self.n_epochs)
                self.log_infos(update_info, self.current_step)
                train_info.update(update_info)
                self.callback.on_train_epochs_end(self.current_step, policy=self.policy, memory=self.memory,
                                                  current_episode=self.current_episode, train_steps=train_steps,
                                                  update_info=update_info)

            scores += rewards
            obs = deepcopy(next_obs)

            if terminal:
                episode_info = {"returns-step": scores}
                scores = 0
                returns = 0
                episodes += 1
                self.end_episode(episodes)
                obs, _ = self.envs.reset()
                self.log_infos(step_info, self.current_step)
                train_info.update(episode_info)
                self.callback.on_train_episode_info(envs=self.envs, policy=self.policy,
                                                    rank=self.rank, use_wandb=self.use_wandb,
                                                    current_step=self.current_step,
                                                    current_episode=self.current_episode,
                                                    train_steps=train_steps)

            self.current_step += self.n_envs

            if self.noise_scale >= self.end_noise:
                self.noise_scale -= self.delta_noise

            self.callback.on_train_step_end(self.current_step, envs=self.envs, policy=self.policy,
                                            train_steps=train_steps, train_info=train_info)

        return train_info



import os
from argparse import Namespace
from copy import deepcopy
import numpy as np
from tqdm import tqdm
from operator import itemgetter
from xuance.torch.agents import REGISTRY_Agents
from xuance.environment import make_envs
from xuance.environment.vector_envs import combine_actions
from xuance.torch.utils.operations import set_seed


class RunnerCompetition(object):
    def __init__(self, configs):
        self.configs = configs
        # set random seeds
        set_seed(self.configs[0].seed)

        # build environments
        self.envs = make_envs(self.configs[0])
        self.n_envs = self.envs.num_envs
        self.current_step = 0
        self.envs.reset()
        self.groups_info = self.envs.groups_info
        self.groups = self.groups_info['agent_groups']
        self.num_groups = self.groups_info['num_groups']
        self.obs_space_groups = self.groups_info['observation_space_groups']
        self.act_space_groups = self.groups_info['action_space_groups']
        assert len(configs) == self.num_groups, "Number of groups must be equal to the number of methods."
        self.agents = []
        for group in range(self.num_groups):
            _env = Namespace(num_agents=len(self.groups[group]),
                             num_envs=self.envs.num_envs,
                             agents=self.groups[group],
                             state_space=self.envs.state_space,
                             observation_space=self.obs_space_groups[group],
                             action_space=self.act_space_groups[group],
                             max_episode_steps=self.envs.max_episode_steps)
            self.agents.append(REGISTRY_Agents[self.configs[group].agent](self.configs[group], _env))

        self.distributed_training = self.agents[0].distributed_training
        self.use_actions_mask = self.agents[0].use_actions_mask
        self.use_global_state = self.agents[0].use_global_state
        self.use_rnn = self.agents[0].use_rnn
        self.use_wandb = self.agents[0].use_wandb

        self.rank = 0
        if self.distributed_training:
            self.rank = int(os.environ['RANK'])

    def rprint(self, info: str):
        if self.rank == 0:
            print(info)

    def run(self):
        if self.configs[0].test_mode:
            def env_fn():
                config_test = deepcopy(self.configs[0])
                config_test.parallels = 1
                config_test.render = True
                return make_envs(config_test)

            for agent in self.agents:
                agent.render = True
                agent.load_model(agent.model_dir_load)

            scores = self.test(env_fn, self.configs[0].test_episode)

            print(f"Mean Score: {scores}, Std: {scores}")
            print("Finish testing.")
        else:
            n_train_steps = self.configs[0].running_steps // self.n_envs

            self.train(n_train_steps)

            print("Finish training.")
            for agent in self.agents:
                agent.save_model("final_train_model.pth")

        for agent in self.agents:
            agent.finish()
        self.envs.close()

    def benchmark(self):
        def env_fn():
            config_test = deepcopy(self.configs[0])
            config_test.parallels = 1  # config_test.test_episode
            return make_envs(config_test)

        train_steps = self.configs[0].running_steps // self.n_envs
        eval_interval = self.configs[0].eval_interval // self.n_envs
        test_episode = self.configs[0].test_episode
        num_epoch = int(train_steps / eval_interval)

        test_scores = self.test(env_fn, test_episode) if self.rank == 0 else 0.0

        best_scores_info = [{"mean": np.mean(test_scores[i]),
                             "std": np.std(test_scores[i]),
                             "step": self.agents[i].current_step} for i in range(self.num_groups)]

        for i_epoch in range(num_epoch):
            print("Epoch: %d/%d:" % (i_epoch, num_epoch))

            self.train(eval_interval)

            if self.rank == 0:

                test_scores = self.test(env_fn, test_episode)

                for i in range(self.num_groups):
                    if np.mean(test_scores[i]) > best_scores_info[i]["mean"]:
                        best_scores_info[i] = {"mean": np.mean(test_scores[i]),
                                               "std": np.std(test_scores[i]),
                                               "step": self.agents[i].current_step}
                        # save best model
                        self.agents[i].save_model(model_name="best_model.pth")

        # end benchmarking
        best_scores = [score["mean"] for score in best_scores_info]
        std_list = [score["std"] for score in best_scores_info]
        print(f"The training for {self.configs[0].env_name}/{self.configs[0].env_id} is finished.")
        print("Algorithms: ", [config.agent for config in self.configs])
        print("Best Model Score: ,", best_scores, "Std: ", std_list)
        for i in range(self.num_groups):
            self.agents[i].finish()
        self.envs.close()

    def train(self, n_steps):
        """
        Train the model for numerous steps.
        Args:
            n_steps (int): Number of steps to train the model:
        """
        if self.use_rnn:
            with tqdm(total=n_steps) as process_bar:
                step_start, step_last = deepcopy(self.current_step), deepcopy(self.current_step)
                n_steps_all = n_steps * self.n_envs
                while step_last - step_start < n_steps_all:
                    self.run_episodes(None, n_episodes=self.n_envs, test_mode=False)
                    for agent in self.agents:
                        if self.current_step >= agent.start_training:
                            train_info = agent.train_epochs(n_epochs=agent.n_epochs)
                            agent.log_infos(train_info, self.current_step)
                    process_bar.update((self.current_step - step_last) // self.n_envs)
                    step_last = deepcopy(self.current_step)
                process_bar.update(n_steps - process_bar.last_print_n)
            return

        obs_dict = self.envs.buf_obs
        avail_actions = self.envs.buf_avail_actions if self.use_actions_mask else None
        state = self.envs.buf_state.copy() if self.use_global_state else None
        for _ in tqdm(range(n_steps)):
            step_info = {}
            policy_out_list = [agent.action(obs_dict=obs_dict,
                                            state=state,
                                            avail_actions_dict=avail_actions,
                                            test_mode=False) for agent in self.agents]
            actions_execute = combine_actions(policy_out_list, self.n_envs)
            next_obs_dict, rewards_dict, terminated_dict, truncated, info = self.envs.step(actions_execute)
            next_state = self.envs.buf_state.copy() if self.use_global_state else None
            next_avail_actions = self.envs.buf_avail_actions.copy() if self.use_actions_mask else None
            for agent in self.agents:
                agent.store_experience(obs_dict, avail_actions, actions_execute, next_obs_dict, next_avail_actions,
                                       rewards_dict, terminated_dict, info,
                                       **{'state': state, 'next_state': next_state})
                if self.current_step >= agent.start_training and self.current_step % agent.training_frequency == 0:
                    train_info = agent.train_epochs(n_epochs=agent.n_epochs)
                    agent.log_infos(train_info, self.current_step)

            obs_dict = next_obs_dict

            if self.use_global_state:
                state = deepcopy(next_state)
            if self.use_actions_mask:
                avail_actions = deepcopy(next_avail_actions)

            for i in range(self.n_envs):
                if all(terminated_dict[i].values()) or truncated[i]:
                    obs_dict[i] = info[i]["reset_obs"]
                    self.envs.buf_obs[i] = info[i]["reset_obs"]
                    if self.use_global_state:
                        state = info[i]["reset_state"]
                        self.envs.buf_state[i] = info[i]["reset_state"]
                    if self.use_actions_mask:
                        avail_actions[i] = info[i]["reset_avail_actions"]
                        self.envs.buf_avail_actions[i] = info[i]["reset_avail_actions"]
                    for agent in self.agents:
                        episode_score = np.mean(itemgetter(*agent.agent_keys)(info[i]["episode_score"]))
                        if self.use_wandb:
                            step_info[f"Train-Results/Episode-Steps/rank_{self.rank}/env-%d" % i] = info[i][
                                "episode_step"]
                            step_info[f"Train-Results/Episode-Rewards/rank_{self.rank}/env-%d" % i] = episode_score
                        else:
                            step_info[f"Train-Results/Episode-Steps/rank_{self.rank}"] = {
                                "env-%d" % i: info[i]["episode_step"]}
                            step_info[f"Train-Results/Episode-Rewards/rank_{self.rank}"] = {"env-%d" % i: episode_score}
                        agent.log_infos(step_info, self.current_step)

            self.current_step += self.n_envs
            for agent in self.agents:
                if not agent.on_policy:
                    agent._update_explore_factor()

    def run_episodes(self, env_fn=None, n_episodes: int = 1, test_mode: bool = False):
        """
        Run some episodes when use RNN.

        Parameters:
            env_fn: The function that can make some testing environments.
            n_episodes (int): Number of episodes.
            test_mode (bool): Whether to test the model.

        Returns:
            Scores: The episode scores.
        """
        envs = self.envs if env_fn is None else env_fn()
        num_envs = envs.num_envs
        videos = [[] for _ in range(num_envs)]
        episode_videos = [[] for _ in range(self.num_groups)]
        episode_count = 0
        scores = [[0.0 for _ in range(num_envs)] for _ in range(self.num_groups)]
        best_score = [-np.inf for _ in range(self.num_groups)]
        obs_dict, info = envs.reset()
        state = envs.buf_state.copy() if self.use_global_state else None
        avail_actions = envs.buf_avail_actions if self.use_actions_mask else None
        if test_mode:
            for config in self.configs:
                if config.render_mode == "rgb_array" and config.render:
                    images = envs.render(config.render_mode)
                    for idx, img in enumerate(images):
                        videos[idx].append(img)
        else:
            if self.use_rnn:
                for agent in self.agents:
                    agent.memory.clear_episodes()
        rnn_hidden = [agent.init_rnn_hidden(num_envs) for agent in self.agents]

        while episode_count < n_episodes:
            step_info = {}
            policy_out_list = [agent.action(obs_dict=obs_dict,
                                            state=state,
                                            avail_actions_dict=avail_actions,
                                            rnn_hidden=rnn_hidden[i_agt],
                                            test_mode=test_mode) for i_agt, agent in enumerate(self.agents)]
            actions_execute = combine_actions(policy_out_list, num_envs)
            rnn_hidden = [policy_out['hidden_state'] for policy_out in policy_out_list]
            next_obs_dict, rewards_dict, terminated_dict, truncated, info = envs.step(actions_execute)
            next_state = envs.buf_state.copy() if self.use_global_state else None
            next_avail_actions = envs.buf_avail_actions if self.use_actions_mask else None
            if test_mode:
                for config in self.configs:
                    if config.render_mode == "rgb_array" and config.render:
                        images = envs.render(config.render_mode)
                        for idx, img in enumerate(images):
                            videos[idx].append(img)
            else:
                for agent in self.agents:
                    agent.store_experience(obs_dict, avail_actions, actions_execute, next_obs_dict, next_avail_actions,
                                           rewards_dict, terminated_dict, info,
                                           **{'state': state, 'next_state': next_state})
            obs_dict = deepcopy(next_obs_dict)
            if self.use_global_state:
                state = deepcopy(next_state)
            if self.use_actions_mask:
                avail_actions = deepcopy(next_avail_actions)

            for i in range(num_envs):
                if all(terminated_dict[i].values()) or truncated[i]:
                    episode_count += 1
                    obs_dict[i] = info[i]["reset_obs"]
                    envs.buf_obs[i] = info[i]["reset_obs"]
                    if self.use_global_state:
                        state = info[i]["reset_state"]
                        self.envs.buf_state[i] = info[i]["reset_state"]
                    if self.use_actions_mask:
                        avail_actions[i] = info[i]["reset_avail_actions"]
                        envs.buf_avail_actions[i] = info[i]["reset_avail_actions"]
                    if self.use_rnn:
                        rnn_hidden = [agent.init_hidden_item(i_env=i, rnn_hidden=rnn_hidden) for agent in self.agents]
                        if not test_mode:
                            terminal_data = {'obs': next_obs_dict[i],
                                             'episode_step': info[i]['episode_step']}
                            if self.use_global_state:
                                terminal_data['state'] = next_state[i]
                            if self.use_actions_mask:
                                terminal_data['avail_actions'] = next_avail_actions[i]
                            for agent in self.agents:
                                agent.memory.finish_path(i, **terminal_data)
                    if not test_mode:
                        self.current_step += info[i]["episode_step"]
                    for i_group in range(self.num_groups):
                        episode_score = float(np.mean(itemgetter(*self.groups[i_group])(info[i]["episode_score"])))
                        scores[i_group].append(episode_score)
                        if test_mode:
                            if best_score[i_group] < episode_score:
                                best_score[i_group] = episode_score
                                episode_videos[i_group] = videos[i].copy()
                            if self.configs[i_group].test_mode:
                                print("Episode: %d, Score: %.2f" % (episode_count, episode_score))
                        else:
                            for agent in self.agents:
                                episode_score = np.mean(itemgetter(*agent.agent_keys)(info[i]["episode_score"]))
                                if self.use_wandb:
                                    step_info["Train-Results/Episode-Steps/env-%d" % i] = info[i]["episode_step"]
                                    step_info["Train-Results/Episode-Rewards/env-%d" % i] = episode_score
                                else:
                                    step_info["Train-Results/Episode-Steps"] = {"env-%d" % i: info[i]["episode_step"]}
                                    step_info["Train-Results/Episode-Rewards"] = {"env-%d" % i: episode_score}
                                agent.log_infos(step_info, self.current_step)
                                if not agent.on_policy:
                                    agent._update_explore_factor()

        if test_mode:
            for i_group in range(self.num_groups):
                config = self.configs[i_group]
                if config.render_mode == "rgb_array" and config.render:
                    # time, height, width, channel -> time, channel, height, width
                    videos_info = {"Videos_Test": np.array([episode_videos[i_group]],
                                                           dtype=np.uint8).transpose((0, 1, 4, 2, 3))}
                    self.agents[i_group].log_videos(info=videos_info, fps=config.fps, x_index=self.current_step)

            if self.configs[0].test_mode:
                print("Best Score: ", best_score)

            for i_group in range(self.num_groups):
                test_info = {
                    "Test-Results/Episode-Rewards": np.mean(scores[i_group]),
                    "Test-Results/Episode-Rewards-Std": np.std(scores[i_group]),
                }

                self.agents[i_group].log_infos(test_info, self.current_step)
            if env_fn is not None:
                envs.close()
        return scores

    def test(self, env_fn, n_episodes):
        """
        Test the model for some episodes.

        Parameters:
            env_fn: The function that can make some testing environments.
            n_episodes (int): Number of episodes to test.

        Returns:
            scores (List(float)): A list of cumulative rewards for each episode.
        """
        scores = self.run_episodes(env_fn=env_fn, n_episodes=n_episodes, test_mode=True)
        return scores



"""
COMA: Counterfactual Multi-Agent Policy Gradients
Paper link: https://ojs.aaai.org/index.php/AAAI/article/view/11794
Implementation: TensorFlow 2.X
"""
from argparse import Namespace
from xuance.common import Sequence
from xuance.tensorflow import tf, tk, Module
from xuance.tensorflow.learners import LearnerMAS


class COMA_Learner(LearnerMAS):
    def __init__(self,
                 config: Namespace,
                 policy: Module,
                 optimizer: Sequence[tk.optimizers.Optimizer],
                 device: str = "cpu:0",
                 model_dir: str = "./",
                 gamma: float = 0.99,
                 sync_frequency: int = 100
                 ):
        self.gamma = gamma
        self.td_lambda = config.td_lambda
        self.sync_frequency = sync_frequency
        self.use_global_state = config.use_global_state
        self.sync_frequency = sync_frequency
        super(COMA_Learner, self).__init__(config, policy, optimizer, device, model_dir)
        self.optimizer = {
            'actor': optimizer[0],
            'critic': optimizer[1]
        }
        self.iterations_actor = self.iterations
        self.iterations_critic = 0

    def update(self, sample, epsilon=0.0):
        self.iterations += 1
        with tf.device(self.device):
            state = tf.convert_to_tensor(sample['state'])
            obs = tf.convert_to_tensor(sample['obs'])
            actions = tf.convert_to_tensor(sample['actions'], dtype=tf.int32)
            actions_onehot = tf.convert_to_tensor(sample['actions_onehot'])
            targets = tf.squeeze(tf.convert_to_tensor(sample['returns']), -1)
            agent_mask = tf.convert_to_tensor(sample['agent_mask'], dtype=tf.float32)
            batch_size = obs.shape[0]
            IDs = tf.tile(tf.expand_dims(tf.eye(self.n_agents), axis=0), multiples=(batch_size, 1, 1))

            with tf.GradientTape() as tape:
                # build critic input
                actions_in = tf.repeat(tf.reshape(tf.expand_dims(actions_onehot, 1), [batch_size, 1, -1]), self.n_agents, 1)
                actions_in_mask = 1 - tf.eye(self.n_agents)
                actions_in_mask = tf.reshape(tf.repeat(tf.reshape(actions_in_mask, [-1, 1]), self.dim_act, 1), [self.n_agents, -1])
                actions_in = actions_in * tf.expand_dims(actions_in_mask, 0)
                if self.use_global_state:
                    state = tf.repeat(tf.expand_dims(state, 1), self.n_agents, 1)
                    critic_in = tf.concat([state, obs, actions_in], axis=-1)
                else:
                    critic_in = tf.concat([obs, actions_in])
                # get critic value
                _, q_eval = self.policy.get_values(critic_in)
                q_eval_a = tf.squeeze(tf.gather(q_eval, tf.expand_dims(actions, -1), axis=-1, batch_dims=-1), -1)
                q_eval_a *= agent_mask
                targets *= agent_mask
                loss_c = tf.reduce_sum((q_eval_a - tf.stop_gradient(targets)) ** 2) / tf.reduce_sum(agent_mask)
                gradients = tape.gradient(loss_c, self.policy.critic_trainable_variables)
                self.optimizer['critic'].apply_gradients([
                    (tf.clip_by_norm(grad, self.args.clip_grad), var)
                    for (grad, var) in zip(gradients, self.policy.critic_trainable_variables)
                    if grad is not None
                ])

            with tf.GradientTape() as tape:
                # calculate baselines
                inputs_policy = {'obs': obs, 'ids': IDs}
                _, pi_probs = self.policy(inputs_policy, epsilon=epsilon)
                baseline = tf.math.reduce_sum(pi_probs * q_eval, axis=-1)
                pi_a = tf.squeeze(tf.gather(pi_probs, tf.expand_dims(actions, -1), axis=-1, batch_dims=-1), -1)
                log_pi_a = tf.math.log(pi_a)
                advantages = tf.stop_gradient(q_eval_a - baseline)
                log_pi_a *= agent_mask
                advantages *= agent_mask
                loss_coma = -tf.reduce_sum(advantages * log_pi_a) / tf.reduce_sum(agent_mask)
                gradients = tape.gradient(loss_coma, self.policy.param_actor())
                self.optimizer['actor'].apply_gradients([
                    (tf.clip_by_norm(grad, self.args.clip_grad), var)
                    for (grad, var) in zip(gradients, self.policy.param_actor())
                    if grad is not None
                ])

            # Logger
            learning_rate_actor = self.optimizer['actor']._decayed_lr(tf.float32)
            learning_rate_critic = self.optimizer['critic']._decayed_lr(tf.float32)

            info = {
                "learning_rate_actor": learning_rate_actor.numpy(),
                "learning_rate_critic": learning_rate_critic.numpy(),
                "actor_loss": loss_coma.numpy(),
                "critic_loss": loss_c.numpy(),
                "advantage": tf.math.reduce_mean(advantages).numpy()
            }

            return info



"""
Multi-Agent Proximal Policy Optimization (MAPPO)
Paper link:
https://arxiv.org/pdf/2103.01955.pdf
Implementation: TensorFlow 2.X
"""
import numpy as np
from argparse import Namespace
from operator import itemgetter
from xuance.common import List
from xuance.tensorflow import tf, tk, Module
from xuance.tensorflow.learners.multi_agent_rl.ippo_learner import IPPO_Learner


class MAPPO_Clip_Learner(IPPO_Learner):
    def __init__(self,
                 config: Namespace,
                 model_keys: List[str],
                 agent_keys: List[str],
                 policy: Module):
        super(MAPPO_Clip_Learner, self).__init__(config, model_keys, agent_keys, policy)

    def update(self, sample):
        self.iterations += 1
        info = {}

        # prepare training data
        sample_Tensor = self.build_training_data(sample=sample,
                                                 use_parameter_sharing=self.use_parameter_sharing,
                                                 use_actions_mask=self.use_actions_mask,
                                                 use_global_state=self.use_global_state)
        batch_size = sample_Tensor['batch_size']
        state = sample_Tensor['state']
        obs = sample_Tensor['obs']
        actions = sample_Tensor['actions']
        agent_mask = sample_Tensor['agent_mask']
        avail_actions = sample_Tensor['avail_actions']
        values = sample_Tensor['values']
        returns = sample_Tensor['returns']
        advantages = sample_Tensor['advantages']
        log_pi_old = sample_Tensor['log_pi_old']
        IDs = sample_Tensor['agent_ids']

        # prepare critic inputs
        if self.use_parameter_sharing:
            key = self.model_keys[0]
            bs = batch_size * self.n_agents
            if self.use_global_state:
                critic_input = {key: state.reshape(batch_size, 1, -1).repeat(self.n_agents, axis=1).reshape(bs, -1)}
            else:
                critic_input = {key: obs[key].reshape(batch_size, 1, -1).repeat(self.n_agents, axis=1).reshape(bs, -1)}
        else:
            bs = batch_size
            if self.use_global_state:
                critic_input = {k: state.reshape(batch_size, -1) for k in self.agent_keys}
            else:
                joint_obs = np.stack(itemgetter(*self.agent_keys)(obs), axis=-1)
                critic_input = {k: joint_obs for k in self.agent_keys}

        with tf.GradientTape() as tape:
            _, pi_dists_dict = self.policy(observation=obs, agent_ids=IDs, avail_actions=avail_actions)
            pi_dists_dict = {key: self.policy.actor[key].dist for key in self.model_keys}
            _, value_pred_dict = self.policy.get_values(observation=critic_input, agent_ids=IDs)

            # calculate losses for each agent
            loss_a, loss_e, loss_c = [], [], []
            for key in self.model_keys:
                mask_values = agent_mask[key]
                # actor loss
                log_pi = tf.reshape(pi_dists_dict[key].log_prob(actions[key]), [bs])
                ratio = tf.reshape(tf.exp(log_pi - log_pi_old[key]), [bs])
                advantages_mask = tf.stop_gradient(advantages[key]) * mask_values
                surrogate1 = ratio * advantages_mask
                surrogate2 = tf.clip_by_value(ratio, 1 - self.clip_range, 1 + self.clip_range) * advantages_mask
                loss_a.append(-tf.reduce_mean(tf.math.minimum(surrogate1, surrogate2)))

                # entropy loss
                entropy = tf.reshape(pi_dists_dict[key].entropy(), [bs]) * mask_values
                loss_e.append(tf.reduce_mean(entropy))

                # critic loss
                value_pred_i = tf.reshape(value_pred_dict[key], [bs])
                value_target = tf.reshape(returns[key], [bs])
                values_i = tf.reshape(values[key], [bs])
                if self.use_value_clip:
                    value_clipped = values_i + tf.clip_by_value(value_pred_i - values_i,
                                                                -self.value_clip_range, self.value_clip_range)
                    if self.use_value_norm:
                        self.value_normalizer[key].update(tf.reshape(value_target, [bs, 1]))
                        value_target = self.value_normalizer[key].normalize(tf.reshape(value_target, [bs, 1]))
                        value_target = tf.reshape(value_target, [bs])
                    if self.use_huber_loss:
                        loss_v = tk.losses.huber(value_target, value_pred_i)
                        loss_v_clipped = tk.losses.huber(value_target, value_clipped)
                    else:
                        loss_v = (value_pred_i - value_target) ** 2
                        loss_v_clipped = (value_clipped - value_target) ** 2
                    loss_c_ = tf.math.maximum(loss_v, loss_v_clipped) * mask_values
                    loss_c.append(tf.reduce_sum(loss_c_) / tf.reduce_sum(mask_values))
                else:
                    if self.use_value_norm:
                        self.value_normalizer[key].update(value_target)
                        value_target = self.value_normalizer[key].normalize(value_target)
                    if self.use_huber_loss:
                        loss_v = tk.losses.huber(value_target, value_pred_i) * mask_values
                    else:
                        loss_v = ((value_pred_i - value_target) ** 2) * mask_values
                    loss_c.append(tf.reduce_sum(loss_v) / tf.reduce_sum(mask_values))

                info.update({f"{key}/actor_loss": loss_a[-1].numpy(),
                             f"{key}/critic_loss": loss_c[-1].numpy(),
                             f"{key}/entropy": loss_e[-1].numpy(),
                             f"{key}/predict_value": tf.reduce_mean(value_pred_i).numpy()})

            loss = sum(loss_a) + self.vf_coef * sum(loss_c) - self.ent_coef * sum(loss_e)

            gradients = tape.gradient(loss, self.policy.trainable_variables)
            if self.use_grad_clip:
                self.optimizer.apply_gradients([
                    (tf.clip_by_norm(grad, self.grad_clip_norm), var)
                    for (grad, var) in zip(gradients, self.policy.trainable_variables)
                    if grad is not None
                ])
            else:
                self.optimizer.apply_gradients([
                    (grad, var)
                    for (grad, var) in zip(gradients, self.policy.trainable_variables)
                    if grad is not None
                ])

            info.update({"loss": loss.numpy()})

        return info



from .pg_agent import PG_Agent
from .ppg_agent import PPG_Agent
from .a2c_agent import A2C_Agent
from .ppokl_agent import PPOKL_Agent
from .ppoclip_agent import PPOCLIP_Agent
from .pdqn_agent import PDQN_Agent
from .spdqn_agent import SPDQN_Agent
from .mpdqn_agent import MPDQN_Agent
from .ddpg_agent import DDPG_Agent
from .sac_agent import SAC_Agent
from .td3_agent import TD3_Agent
from .npg_agent import NPG_Agent



from tqdm import tqdm
import numpy as np
from copy import deepcopy
from argparse import Namespace
from operator import itemgetter
from xuance.common import Optional, List, Union, MARL_OffPolicyBuffer, MARL_OffPolicyBuffer_RNN
from xuance.environment import DummyVecMultiAgentEnv, SubprocVecMultiAgentEnv
from xuance.tensorflow import Tensor, Module
from xuance.tensorflow.utils.distributions import Categorical
from xuance.tensorflow.agents.base import MARLAgents


class OffPolicyMARLAgents(MARLAgents):
    """The core class for off-policy algorithm with multiple agents.

    Args:
        config: the Namespace variable that provides hyperparameters and other settings.
        envs: the vectorized environments.
        callback: A user-defined callback function object to inject custom logic during training.
    """

    def __init__(self,
                 config: Namespace,
                 envs: Union[DummyVecMultiAgentEnv, SubprocVecMultiAgentEnv]):
        super(OffPolicyMARLAgents, self).__init__(config, envs)
        self.start_greedy = config.start_greedy if hasattr(config, "start_greedy") else None
        self.end_greedy = config.end_greedy if hasattr(config, "start_greedy") else None
        self.delta_egreedy: Optional[float] = None
        self.e_greedy: Optional[float] = None

        self.start_noise = config.start_noise if hasattr(config, "start_noise") else None
        self.end_noise = config.end_noise if hasattr(config, "end_noise") else None
        self.delta_noise: Optional[float] = None
        self.noise_scale: Optional[float] = None
        self.actions_low = self.action_space.low if hasattr(self.action_space, "low") else None
        self.actions_high = self.action_space.high if hasattr(self.action_space, "high") else None

        self.auxiliary_info_shape = None
        self.memory: Optional[MARL_OffPolicyBuffer, MARL_OffPolicyBuffer_RNN] = None

    def _build_memory(self):
        """Build replay buffer for models training
        """
        if self.use_actions_mask:
            avail_actions_shape = {key: (self.action_space[key].n,) for key in self.agent_keys}
        else:
            avail_actions_shape = None
        input_buffer = dict(agent_keys=self.agent_keys,
                            state_space=self.state_space if self.use_global_state else None,
                            obs_space=self.observation_space,
                            act_space=self.action_space,
                            n_envs=self.n_envs,
                            buffer_size=self.config.buffer_size,
                            batch_size=self.config.batch_size,
                            avail_actions_shape=avail_actions_shape,
                            use_actions_mask=self.use_actions_mask,
                            max_episode_steps=self.episode_length)
        Buffer = MARL_OffPolicyBuffer_RNN if self.use_rnn else MARL_OffPolicyBuffer
        return Buffer(**input_buffer)

    def _build_policy(self) -> Module:
        raise NotImplementedError

    def store_experience(self, obs_dict, avail_actions, actions_dict, obs_next_dict, avail_actions_next,
                         rewards_dict, terminals_dict, info, **kwargs):
        """
        Store experience data into replay buffer.

        Parameters:
            obs_dict (List[dict]): Observations for each agent in self.agent_keys.
            avail_actions (List[dict]): Actions mask values for each agent in self.agent_keys.
            actions_dict (List[dict]): Actions for each agent in self.agent_keys.
            obs_next_dict (List[dict]): Next observations for each agent in self.agent_keys.
            avail_actions_next (List[dict]): The next actions mask values for each agent in self.agent_keys.
            rewards_dict (List[dict]): Rewards for each agent in self.agent_keys.
            terminals_dict (List[dict]): Terminated values for each agent in self.agent_keys.
            info (List[dict]): Other information for the environment at current step.
        """
        experience_data = {
            'obs': {k: np.array([data[k] for data in obs_dict]) for k in self.agent_keys},
            'actions': {k: np.array([data[k] for data in actions_dict]) for k in self.agent_keys},
            'obs_next': {k: np.array([data[k] for data in obs_next_dict]) for k in self.agent_keys},
            'rewards': {k: np.array([data[k] for data in rewards_dict]) for k in self.agent_keys},
            'terminals': {k: np.array([data[k] for data in terminals_dict]) for k in self.agent_keys},
            'agent_mask': {k: np.array([data['agent_mask'][k] for data in info]) for k in self.agent_keys},
        }
        if self.use_rnn:
            experience_data['episode_steps'] = np.array([data['episode_step'] - 1 for data in info])
        if self.use_global_state:
            experience_data['state'] = np.array(kwargs['state'])
            experience_data['state_next'] = np.array(kwargs['next_state'])
        if self.use_actions_mask:
            experience_data['avail_actions'] = {k: np.array([data[k] for data in avail_actions])
                                                for k in self.agent_keys}
            experience_data['avail_actions_next'] = {k: np.array([data[k] for data in avail_actions_next])
                                                     for k in self.agent_keys}
        self.memory.store(**experience_data)

    def init_rnn_hidden(self, n_envs):
        """
        Returns initialized hidden states of RNN if use RNN-based representations.

        Parameters:
            n_envs (int): The number of parallel environments.

        Returns:
            rnn_hidden_states: The hidden states for RNN.
        """
        rnn_hidden_states = None
        if self.use_rnn:
            batch = n_envs * self.n_agents if self.use_parameter_sharing else n_envs
            rnn_hidden_states = {k: self.policy.representation[k].init_hidden(batch) for k in self.model_keys}
        return rnn_hidden_states

    def init_hidden_item(self, i_env: int,
                         rnn_hidden: Optional[dict] = None):
        """
        Returns initialized hidden states of RNN for i-th environment.

        Parameters:
            i_env (int): The index of environment that to be selected.
            rnn_hidden (Optional[dict]): The RNN hidden states of actor representation.
        """
        assert self.use_rnn is True, "This method cannot be called when self.use_rnn is False."
        if self.use_parameter_sharing:
            batch_index = list(range(i_env * self.n_agents, (i_env + 1) * self.n_agents))
        else:
            batch_index = [i_env, ]
        for key in self.model_keys:
            rnn_hidden[key] = self.policy.representation[key].init_hidden_item(batch_index, *rnn_hidden[key])
        return rnn_hidden

    def _update_explore_factor(self):
        if self.e_greedy is not None:
            if self.e_greedy > self.end_greedy:
                self.e_greedy = self.start_greedy - self.delta_egreedy * self.current_step
            else:
                self.e_greedy = self.end_greedy
        elif self.noise_scale is not None:
            if self.noise_scale >= self.end_noise:
                self.noise_scale = self.start_noise - self.delta_noise * self.current_step
            else:
                self.noise_scale = self.end_noise
        else:
            return

    def exploration(self, batch_size: int,
                    pi_actions_dict: Union[List[dict], dict],
                    avail_actions_dict: Optional[List[dict]] = None):
        """Returns the actions for exploration.

        Parameters:
            batch_size (int): The batch size.
            pi_actions_dict (Optional[List[dict], dict]): The original output actions.
            avail_actions_dict (Optional[List[dict]]): Actions mask values, default is None.

        Returns:
            explore_actions: The actions with noisy values.
        """
        if self.e_greedy is not None:
            if np.random.rand() < self.e_greedy:
                if self.use_actions_mask:
                    explore_actions = [{k: Categorical(Tensor(avail_actions_dict[e][k])).sample().numpy()
                                        for k in self.agent_keys} for e in range(batch_size)]
                else:
                    explore_actions = [{k: self.action_space[k].sample() for k in self.agent_keys} for _ in
                                       range(batch_size)]
            else:
                explore_actions = pi_actions_dict
        elif self.noise_scale is not None:
            if self.use_parameter_sharing:
                key = self.agent_keys[0]
                pi_actions_dict[key] += np.random.normal(0, self.noise_scale, size=pi_actions_dict[key].shape)
            else:
                for key in self.agent_keys:
                    pi_actions_dict[key] += np.random.normal(0, self.noise_scale, size=pi_actions_dict[key].shape)
            explore_actions = pi_actions_dict
        else:
            explore_actions = pi_actions_dict
        return explore_actions

    def action(self,
               obs_dict: List[dict],
               avail_actions_dict: Optional[List[dict]] = None,
               rnn_hidden: Optional[dict] = None,
               test_mode: Optional[bool] = False):
        """
        Returns actions for agents.

        Parameters:
            obs_dict (List[dict]): Observations for each agent in self.agent_keys.
            avail_actions_dict (Optional[List[dict]]): Actions mask values, default is None.
            rnn_hidden (Optional[dict]): The hidden variables of the RNN.
            test_mode (Optional[bool]): True for testing without noises.

        Returns:
            rnn_hidden_state (dict): The new hidden states for RNN (if self.use_rnn=True).
            actions_dict (dict): The output actions.
        """
        batch_size = len(obs_dict)
        obs_input, agents_id, avail_actions_input = self._build_inputs(obs_dict, avail_actions_dict)
        hidden_state, actions, _ = self.policy(observation=obs_input,
                                               agent_ids=agents_id,
                                               avail_actions=avail_actions_input,
                                               rnn_hidden=rnn_hidden)

        if self.use_parameter_sharing:
            key = self.agent_keys[0]
            actions_out = actions[key].numpy().reshape([batch_size, self.n_agents])
            actions_dict = [{k: actions_out[e, i] for i, k in enumerate(self.agent_keys)} for e in range(batch_size)]
        else:
            actions_out = {k: actions[k].numpy().reshape(batch_size) for k in self.agent_keys}
            actions_dict = [{k: actions_out[k][i] for k in self.agent_keys} for i in range(batch_size)]

        if not test_mode:  # get random actions
            actions_dict = self.exploration(batch_size, actions_dict, avail_actions_dict)
        return {"hidden_state": hidden_state, "actions": actions_dict}

    def train(self, n_steps):
        """
        Train the model for numerous steps.

        Parameters:
            n_steps (int): The number of steps to train the model.
        """
        if self.use_rnn:
            with tqdm(total=n_steps) as process_bar:
                step_start, step_last = deepcopy(self.current_step), deepcopy(self.current_step)
                n_steps_all = n_steps * self.n_envs
                while step_last - step_start < n_steps_all:
                    self.run_episodes(None, n_episodes=self.n_envs, test_mode=False)
                    if self.current_step >= self.start_training:
                        train_info = self.train_epochs(n_epochs=self.n_epochs)
                        self.log_infos(train_info, self.current_step)
                    process_bar.update((self.current_step - step_last) // self.n_envs)
                    step_last = deepcopy(self.current_step)
                process_bar.update(n_steps - process_bar.last_print_n)
            return

        obs_dict = self.envs.buf_obs
        avail_actions = self.envs.buf_avail_actions if self.use_actions_mask else None
        state = self.envs.buf_state.copy() if self.use_global_state else None
        for _ in tqdm(range(n_steps)):
            step_info = {}
            policy_out = self.action(obs_dict=obs_dict, avail_actions_dict=avail_actions, test_mode=False)
            actions_dict = policy_out['actions']
            next_obs_dict, rewards_dict, terminated_dict, truncated, info = self.envs.step(actions_dict)
            next_state = self.envs.buf_state.copy() if self.use_global_state else None
            next_avail_actions = self.envs.buf_avail_actions if self.use_actions_mask else None
            self.store_experience(obs_dict, avail_actions, actions_dict, next_obs_dict, next_avail_actions,
                                  rewards_dict, terminated_dict, info,
                                  **{'state': state, 'next_state': next_state})
            if self.current_step >= self.start_training and self.current_step % self.training_frequency == 0:
                train_info = self.train_epochs(n_epochs=self.n_epochs)
                self.log_infos(train_info, self.current_step)
            obs_dict = deepcopy(next_obs_dict)
            if self.use_global_state:
                state = deepcopy(next_state)
            if self.use_actions_mask:
                avail_actions = deepcopy(next_avail_actions)

            for i in range(self.n_envs):
                if all(terminated_dict[i].values()) or truncated[i]:
                    obs_dict[i] = info[i]["reset_obs"]
                    self.envs.buf_obs[i] = info[i]["reset_obs"]
                    if self.use_global_state:
                        state = info[i]["reset_state"]
                        self.envs.buf_state[i] = info[i]["reset_state"]
                    if self.use_actions_mask:
                        avail_actions[i] = info[i]["reset_avail_actions"]
                        self.envs.buf_avail_actions[i] = info[i]["reset_avail_actions"]
                    if self.use_wandb:
                        step_info["Train-Results/Episode-Steps/env-%d" % i] = info[i]["episode_step"]
                        step_info["Train-Results/Episode-Rewards/env-%d" % i] = info[i]["episode_score"]
                    else:
                        step_info["Train-Results/Episode-Steps"] = {"env-%d" % i: info[i]["episode_step"]}
                        step_info["Train-Results/Episode-Rewards"] = {
                            "env-%d" % i: np.mean(itemgetter(*self.agent_keys)(info[i]["episode_score"]))}
                    self.log_infos(step_info, self.current_step)

            self.current_step += self.n_envs
            self._update_explore_factor()

    def run_episodes(self, env_fn=None, n_episodes: int = 1, test_mode: bool = False):
        """
        Run some episodes when use RNN.

        Parameters:
            env_fn: The function that can make some testing environments.
            n_episodes (int): Number of episodes.
            test_mode (bool): Whether to test the model.

        Returns:
            Scores: The episode scores.
        """
        envs = self.envs if env_fn is None else env_fn()
        num_envs = envs.num_envs
        videos, episode_videos = [[] for _ in range(num_envs)], []
        episode_count, scores, best_score = 0, [0.0 for _ in range(num_envs)], -np.inf
        obs_dict, info = envs.reset()
        state = envs.buf_state.copy() if self.use_global_state else None
        avail_actions = envs.buf_avail_actions if self.use_actions_mask else None
        if test_mode:
            if self.config.render_mode == "rgb_array" and self.render:
                images = envs.render(self.config.render_mode)
                for idx, img in enumerate(images):
                    videos[idx].append(img)
        else:
            if self.use_rnn:
                self.memory.clear_episodes()
        rnn_hidden = self.init_rnn_hidden(num_envs)

        while episode_count < n_episodes:
            step_info = {}
            policy_out = self.action(obs_dict=obs_dict,
                                     avail_actions_dict=avail_actions,
                                     rnn_hidden=rnn_hidden,
                                     test_mode=test_mode)
            rnn_hidden, actions_dict = policy_out['hidden_state'], policy_out['actions']
            next_obs_dict, rewards_dict, terminated_dict, truncated, info = envs.step(actions_dict)
            next_state = envs.buf_state.copy() if self.use_global_state else None
            next_avail_actions = envs.buf_avail_actions if self.use_actions_mask else None
            if test_mode:
                if self.config.render_mode == "rgb_array" and self.render:
                    images = envs.render(self.config.render_mode)
                    for idx, img in enumerate(images):
                        videos[idx].append(img)
            else:
                self.store_experience(obs_dict, avail_actions, actions_dict, next_obs_dict, next_avail_actions,
                                      rewards_dict, terminated_dict, info,
                                      **{'state': state, 'next_state': next_state})
            obs_dict = deepcopy(next_obs_dict)
            if self.use_global_state:
                state = deepcopy(next_state)
            if self.use_actions_mask:
                avail_actions = deepcopy(next_avail_actions)

            for i in range(num_envs):
                if all(terminated_dict[i].values()) or truncated[i]:
                    episode_count += 1
                    obs_dict[i] = info[i]["reset_obs"]
                    envs.buf_obs[i] = info[i]["reset_obs"]
                    if self.use_global_state:
                        state = info[i]["reset_state"]
                        self.envs.buf_state[i] = info[i]["reset_state"]
                    if self.use_actions_mask:
                        avail_actions[i] = info[i]["reset_avail_actions"]
                        envs.buf_avail_actions[i] = info[i]["reset_avail_actions"]
                    if self.use_rnn:
                        rnn_hidden = self.init_hidden_item(i_env=i, rnn_hidden=rnn_hidden)
                        if not test_mode:
                            terminal_data = {'obs': next_obs_dict[i],
                                             'episode_step': info[i]['episode_step']}
                            if self.use_global_state:
                                terminal_data['state'] = next_state[i]
                            if self.use_actions_mask:
                                terminal_data['avail_actions'] = next_avail_actions[i]
                            self.memory.finish_path(i, **terminal_data)
                    episode_score = float(np.mean(itemgetter(*self.agent_keys)(info[i]["episode_score"])))
                    scores.append(episode_score)
                    if test_mode:
                        if best_score < episode_score:
                            best_score = episode_score
                            episode_videos = videos[i].copy()
                        if self.config.test_mode:
                            print("Episode: %d, Score: %.2f" % (episode_count, episode_score))
                    else:
                        if self.use_wandb:
                            step_info["Train-Results/Episode-Steps/env-%d" % i] = info[i]["episode_step"]
                            step_info["Train-Results/Episode-Rewards/env-%d" % i] = info[i]["episode_score"]
                        else:
                            step_info["Train-Results/Episode-Steps"] = {"env-%d" % i: info[i]["episode_step"]}
                            step_info["Train-Results/Episode-Rewards"] = {
                                "env-%d" % i: np.mean(itemgetter(*self.agent_keys)(info[i]["episode_score"]))}
                        self.current_step += info[i]["episode_step"]
                        self.log_infos(step_info, self.current_step)
                        self._update_explore_factor()

        if test_mode:
            if self.config.render_mode == "rgb_array" and self.render:
                # time, height, width, channel -> time, channel, height, width
                videos_info = {"Videos_Test": np.array([episode_videos], dtype=np.uint8).transpose((0, 1, 4, 2, 3))}
                self.log_videos(info=videos_info, fps=self.fps, x_index=self.current_step)

            if self.config.test_mode:
                print("Best Score: %.2f" % best_score)

            test_info = {
                "Test-Results/Episode-Rewards": np.mean(scores),
                "Test-Results/Episode-Rewards-Std": np.std(scores),
            }

            self.log_infos(test_info, self.current_step)
            if env_fn is not None:
                envs.close()
        return scores

    def train_epochs(self, n_epochs=1):
        """
        Train the model for numerous epochs.

        Parameters:
            n_epochs (int): The number of epochs to train.

        Returns:
            info_train (dict): The information of training.
        """
        info_train = {}
        for i_epoch in range(n_epochs):
            sample = self.memory.sample()
            if self.use_rnn:
                info_train = self.learner.update_rnn(sample)
            else:
                info_train = self.learner.update(sample)
        info_train["epsilon-greedy"] = self.e_greedy
        info_train["noise_scale"] = self.noise_scale
        return info_train

    def test(self, env_fn, n_episodes):
        """
        Test the model for some episodes.

        Parameters:
            env_fn: The function that can make some testing environments.
            n_episodes (int): Number of episodes to test.

        Returns:
            scores (List(float)): A list of cumulative rewards for each episode.
        """
        scores = self.run_episodes(env_fn=env_fn, n_episodes=n_episodes, test_mode=True)
        return scores



import optuna
import importlib
from copy import deepcopy
from argparse import Namespace
from operator import itemgetter
from xuance.environment import make_envs
from xuance.common import get_configs, Optional, List
from xuance.common.tuning_tools.hyperparameters import Hyperparameter, AlgorithmHyperparametersRegistry


def build_search_space(trial: optuna.trial, hyperparameters: List[Hyperparameter]) -> dict:
    '''
    Build the search space for tuning hyperparameters using Optuna.

    This function iterates over a list of hyperparameters and defines the search space
    based on their type and distribution. It supports categorical, float (both uniform
    and log-uniform), and integer hyperparameters.

    Args:
        trial (optuna.trial.Trial): The Optuna trial object used to suggest hyperparameter values.
        hyperparameters (List[Hyperparameter]): A list of Hyperparameter instances defining
            the hyperparameters to tune.

    Returns:
        dict: A dictionary mapping hyperparameter names to their suggested values for the current trial.

    Raises:
        ValueError: If an unsupported hyperparameter type is encountered.

    Example:
        >>> hyperparams = [
        ...     Hyperparameter(name="learning_rate", type="float", distribution=(1e-5, 1e-2), log=True, default=1e-3),
        ...     Hyperparameter(name="batch_size", type="int", distribution=[32, 64, 128, 256], default=64),
        ... ]
        >>> trial = optuna.trial.create_trial()
        >>> search_space = build_search_space(trial, hyperparams)
    '''
    search_space = {}
    for param in hyperparameters:
        if isinstance(param.distribution, list) or param.type == "categorical":
            search_space[param.name] = trial.suggest_categorical(param.name, param.distribution)
        else:
            if param.type == "float":
                if param.log:
                    search_space[param.name] = trial.suggest_loguniform(param.name,
                                                                        low=param.distribution[0],
                                                                        high=param.distribution[1])
                else:
                    search_space[param.name] = trial.suggest_uniform(param.name,
                                                                     low=param.distribution[0],
                                                                     high=param.distribution[1])
            elif param.type == "int":
                search_space[param.name] = trial.suggest_int(param.name, param.distribution[0], param.distribution[1])
            else:
                raise ValueError(f"Unsupported hyperparameter type: {param.type}")
    return search_space


def set_hyperparameters(hyperparameters: List[Hyperparameter], overrides: dict) -> List[Hyperparameter]:
    '''
    Override the distributions of specified hyperparameters.

    This function updates the distribution of hyperparameters based on the provided overrides.
    It ensures that the new distribution is either a tuple or a list. If an unsupported
    distribution type is provided, it raises a ValueError.

    Args:
        hyperparameters (List[Hyperparameter]): The list of Hyperparameter instances to override.
        overrides (dict): A dictionary mapping hyperparameter names to their new distributions.

    Returns:
        List[Hyperparameter]: The updated list of Hyperparameter instances with overridden distributions.

    Raises:
        ValueError: If an unsupported distribution type is provided for any hyperparameter.

    Example:
        >>> hyperparams = [
        ...     Hyperparameter(name="learning_rate", type="float", distribution=(1e-5, 1e-2), log=True, default=1e-3),
        ...     Hyperparameter(name="batch_size", type="int", distribution=[32, 64, 128, 256], default=64),
        ... ]
        >>> overrides = {
        ...     "learning_rate": [0.001, 0.0001, 0.00001],
        ... }
        >>> updated_hyperparams = set_hyperparameters(hyperparams, overrides)
    '''
    for param in hyperparameters:
        if param.name in overrides.keys():
            new_distribution = overrides[param.name]
            if isinstance(new_distribution, tuple) or isinstance(new_distribution, list):
                param.distribution = new_distribution
            else:
                raise ValueError(f"Unsupported distribution type for {param.name}: {type(new_distribution)}")
    return hyperparameters


class HyperParameterTuner:
    """
    A class to facilitate automatic hyperparameter tuning for reinforcement learning algorithms using Optuna.

    The HyperParameterTuner class provides methods to list, select, and tune hyperparameters for a specified
    algorithm within the XuanCe framework. It integrates with Optuna to perform efficient hyperparameter
    optimization.

    Attributes:
        method (str): The name of the method or agent (e.g., 'dqn').
        config_path (str): The path to the configuration YAML file.
        running_steps (Optional[int]): Number of steps to run a trial. Defaults to the value in the configuration.
        test_episodes (Optional[int]): Number of episodes to evaluate the agent's policy. Defaults to the value in the configuration.
        agent_name (str): The name of the agent as specified in the configuration.
        agent: The agent instance retrieved from the agent registry.
    """
    def __init__(self,
                 method: str,
                 config_path: str,
                 running_steps: Optional[int] = None,
                 test_episodes: Optional[int] = None):
        """
        Initialize the HyperParameterTuner module.

        This constructor sets up the tuner by loading configurations, selecting the appropriate
        deep learning toolbox, and registering the algorithm's hyperparameters.

        Args:
            method (str): The name of the method or agent (e.g., 'dqn').
            config_path (str): The path to the configuration YAML file.
            running_steps (Optional[int], optional): Number of steps to run a trial. Defaults to None,
                which means it will use the value from the configuration.
            test_episodes (Optional[int], optional): Number of episodes to evaluate the agent's policy.
                Defaults to None, which means it will use the value from the configuration.

        Raises:
            AttributeError: If the specified deep learning toolbox is not supported.
        """
        self.method = method
        self.configs_dict = get_configs(config_path)
        self.running_steps = self.configs_dict['running_steps'] if running_steps is None else running_steps
        self.test_episodes = self.configs_dict['test_episodes'] if test_episodes is None else test_episodes
        if self.configs_dict['dl_toolbox'] == "torch":
            from xuance.torch.agents import REGISTRY_Agents
        elif self.configs_dict['dl_toolbox'] == "tensorflow":
            from xuance.tensorflow.agents import REGISTRY_Agents
        elif self.configs_dict['dl_toolbox'] == "mindspore":
            from xuance.mindspore.agents import REGISTRY_Agents
        else:
            raise AttributeError(f"XuanCe currently does not support {self.configs_dict['dl_toolbox']}!")
        self.agent_name = self.configs_dict['agent']
        self.agent = REGISTRY_Agents[self.agent_name]
        module = importlib.import_module(f"xuance.common.tuning_tools.hyperparameters.{self.method}")
        params = getattr(module, f"{self.method}_hyperparams")
        AlgorithmHyperparametersRegistry.register_algorithm(self.configs_dict['agent'], params)

    def list_hyperparameters(self) -> List[Hyperparameter]:
        """
        List the hyperparameters of the selected algorithm.

        This method retrieves all registered hyperparameters for the algorithm specified during
        the initialization of the tuner.

        Returns:
            List[Hyperparameter]: A list of Hyperparameter instances associated with the selected algorithm.

        Example:
            >>> tuner = HyperParameterTuner(method='dqn', config_path='config.yaml')
            >>> hyperparams = tuner.list_hyperparameters()
            >>> for hp in hyperparams:
            ...     print(hp.name, hp.type)
            learning_rate float
            gamma float
        """
        return AlgorithmHyperparametersRegistry.get_hyperparameters(self.agent_name)

    def select_hyperparameter(self, hyperparameter_names: List[str]) -> List[Hyperparameter]:
        """
        Select specific hyperparameters for tuning based on their names.

        This method filters the list of all hyperparameters to include only those specified
        by the user. It raises an error if no hyperparameters are selected.

        Args:
            hyperparameter_names (List[str]): A list of hyperparameter names to select for tuning.

        Returns:
            List[Hyperparameter]: A list of selected Hyperparameter instances.

        Raises:
            ValueError: If no hyperparameters are selected for tuning.

        Example:
            >>> tuner = HyperParameterTuner(method='dqn', config_path='config.yaml')
            >>> selected = tuner.select_hyperparameter(['learning_rate', 'gamma'])
            >>> for hp in selected:
            ...     print(hp.name, hp.type)
            learning_rate float
            gamma float
        """
        all_hyperparams = self.list_hyperparameters()
        selected_hyperparams = [param for param in all_hyperparams if param.name in hyperparameter_names]
        if not selected_hyperparams:
            raise ValueError("No hyperparameters selected for tuning.")
        return selected_hyperparams

    def eval_env_fn(self):
        """
        Create the environment for evaluating the agent's policy.

        This method configures a single (vectorized) environment instance used solely for
        evaluating the performance of the trained agent.

        Returns:
            Vectorized Environment: An instance of the environment configured for evaluation.
        """
        configs_test = Namespace(**self.configs_dict)
        configs_test.parallels = 1
        return make_envs(configs_test)

    def objective(self, trail: optuna.trial, selected_hyperparameters: List[Hyperparameter]) -> float:
        """
        Define the objective function for Optuna optimization.

        This function builds the search space, updates the configuration with suggested
        hyperparameters, initializes the environment and agent, trains the agent, evaluates
        its performance, and returns the mean score as the objective value.

        Args:
            trial (optuna.trial.Trial): The Optuna trial object used for suggesting hyperparameter values.
            selected_hyperparameters (List[Hyperparameter]): A list of Hyperparameter instances selected for tuning.

        Returns:
            float: The mean score obtained from evaluating the agent's policy.

        Example:
            >>> tuner = HyperParameterTuner(method='dqn', config_path='config.yaml')
            >>> hyperparams = tuner.select_hyperparameter(['learning_rate', 'gamma'])
            >>> study = optuna.create_study(direction="maximize")
            >>> study.optimize(lambda trial: tuner.objective(trial, hyperparams), n_trials=10)
        """
        search_space = build_search_space(trail, selected_hyperparameters)
        config_trail = deepcopy(self.configs_dict)
        config_trail.update(search_space)
        configs_trail = Namespace(**config_trail)
        envs_trail = make_envs(configs_trail)
        agent_trail = self.agent(configs_trail, envs_trail)
        agent_trail.train(train_steps=self.running_steps)
        scores = agent_trail.test(env_fn=self.eval_env_fn, test_episodes=self.test_episodes)
        agent_trail.finish()
        envs_trail.close()
        scores_mean = sum(scores) / len(scores)
        return scores_mean

    def tune(self,
             selected_hyperparameters: List[Hyperparameter],
             n_trials: int = 1,
             pruner: Optional[optuna.pruners.BasePruner] = None,
             direction: str = "maximize") -> optuna.study.Study:
        """
        Start the hyperparameter tuning process.

        This method initializes an Optuna study, defines the objective function wrapper,
        and begins the optimization process to search for the best hyperparameter values.

        Args:
            selected_hyperparameters (List[Hyperparameter]): A list of Hyperparameter instances selected for tuning.
            n_trials (int, optional): The number of trials to run during optimization. Defaults to 1.
            pruner (Optional[optuna.pruners.BasePruner], optional): An Optuna pruner to terminate unpromising trials early.
                Defaults to None.
            direction (str): The optimization directions. Defaults to "maximize".

        Returns:
            optuna.study.Study: The Optuna study object containing the results of the optimization.

        Example:
            >>> tuner = HyperParameterTuner(method='dqn', config_path='config.yaml')
            >>> hyperparams = tuner.select_hyperparameter(['learning_rate', 'gamma'])
            >>> study = tuner.tune(selected_hyperparameters=hyperparams, n_trials=30)
            >>> print(study.best_params)
        """
        study = optuna.create_study(direction=direction, pruner=pruner)

        def objective_wrapper(trial):
            return self.objective(trial, selected_hyperparameters)

        study.optimize(objective_wrapper, n_trials=n_trials)

        print("Best hyperparameters: ", study.best_params)
        print("Best value: ", study.best_value)

        return study


class MultiObjectiveTuner(HyperParameterTuner):
    """
    A class to facilitate multi-objective hyperparameter tuning for reinforcement learning algorithms using Optuna.
    """
    def __init__(self, **kwargs):
        super(MultiObjectiveTuner, self).__init__(**kwargs)

    def objective(self, trial: optuna.trial, selected_hyperparameters: List[Hyperparameter],
                  selected_objectives: List[str] = None) -> float:
        """
        Define the objective function for Optuna optimization.

        This function builds the search space, updates the configuration with suggested
        hyperparameters, initializes the environment and agent, trains the agent, evaluates
        its performance, and returns the mean score as the objective value.

        Args:
            trial (optuna.trial.Trial): The Optuna trial object used for suggesting hyperparameter values.
            selected_hyperparameters (List[Hyperparameter]): A list of Hyperparameter instances selected for tuning.
            selected_objectives (List[str]): A list of objectives selected for tuning.

        Returns:
            float: The mean score obtained from evaluating the agent's policy.

        Example:
            >>> tuner = HyperParameterTuner(method='dqn', config_path='config.yaml')
            >>> hyperparams = tuner.select_hyperparameter(['learning_rate', 'gamma'])
            >>> study = optuna.create_study(direction="maximize")
            >>> study.optimize(lambda trial: tuner.objective(trial, hyperparams), n_trials=10)
        """
        search_space = build_search_space(trial, selected_hyperparameters)
        config_trail = deepcopy(self.configs_dict)
        config_trail.update(search_space)
        configs_trail = Namespace(**config_trail)
        envs_trail = make_envs(configs_trail)
        agent_trail = self.agent(configs_trail, envs_trail)
        train_info = agent_trail.train(train_steps=self.running_steps)
        scores = agent_trail.test(env_fn=self.eval_env_fn, test_episodes=self.test_episodes)
        agent_trail.finish()
        envs_trail.close()
        scores_mean = sum(scores) / len(scores)
        train_info["test_score"] = scores_mean
        objectives = itemgetter(*selected_objectives)(train_info)
        return objectives

    def tune(self,
             selected_hyperparameters: List[Hyperparameter],
             n_trials: int = 1,
             pruner: Optional[optuna.pruners.BasePruner] = None,
             directions: Optional[list] = None,
             selected_objectives: List[str] = None,) -> optuna.study.Study:
        """
        Start the hyperparameter tuning process.

        This method initializes an Optuna study, defines the objective function wrapper,
        and begins the optimization process to search for the best hyperparameter values.

        Args:
            selected_hyperparameters (List[Hyperparameter]): A list of Hyperparameter instances selected for tuning.
            n_trials (int, optional): The number of trials to run during optimization. Defaults to 1.
            pruner (Optional[optuna.pruners.BasePruner], optional): An Optuna pruner to terminate unpromising trials early.
                Defaults to None.
            directions: The optimization directions. Default is None.
            selected_objectives (List[str]): A list of objectives selected for tuning.

        Returns:
            optuna.study.Study: The Optuna study object containing the results of the optimization.

        Example:
            >>> tuner = HyperParameterTuner(method='dqn', config_path='config.yaml')
            >>> hyperparams = tuner.select_hyperparameter(['learning_rate', 'gamma'])
            >>> study = tuner.tune(selected_hyperparameters=hyperparams, n_trials=30, selected_objectives=['test_score', 'loss'])
            >>> print(study.best_params)
        """
        study = optuna.create_study(directions=directions, pruner=pruner)

        def objective_wrapper(trial):
            return self.objective(trial, selected_hyperparameters, selected_objectives)

        study.optimize(objective_wrapper, n_trials=n_trials)

        print("Number of finished trials: ", len(study.trials))

        return study
